# Scoop.pm -- main sitescooper logic.

package Scoop;

require Exporter;

use Carp;
use File::Find;
use File::Path;
use File::Copy;
use File::Basename;
use Cwd;
use FindBin;

use strict;
use vars qw{
	@ISA @EXPORT $VERSION
	$match_url_dd $match_url_mm $match_url_yyyy $match_url_Mstr
	$match_url_yy
	$MY_OS 
};

@ISA = qw(Exporter);
@EXPORT= qw();
$VERSION = "3.0.1";
sub Version { $VERSION; }

sub new {
  my $class = shift; $class = ref($class) || $class;

  my $self = {
    'httpclient_queue'		=> { },
    'httpclient_queue_new_idx'	=> 0,
  };

  bless ($self, $class);
  $self->LoadStatics();
  $self->Init();
  $self;
}

sub LoadStatics {
  my $self = shift;

  if (!defined $Scoop::__loaded_statics) {
    $Scoop::__loaded_statics = 1;	 # only once
  } else {
    return;
  }

  $Scoop::OUT_TEXT = 0;
  $Scoop::OUT_DOC = 1;
  $Scoop::OUT_HTML = 2;

  $Scoop::home_url = "http://sitescooper.org";
  $Scoop::bookmark_char = "\x8D";	# yes, same as Chris' one, cheers!

# Andrew Fletcher <fletch@computer.org>:
# A relative path on Mac seems to need a ":" before it. I've called
# this $CF::colon. (jm note: currently unused but we may need it again)

  if (&Scoop::MyOS eq 'UNIX') {
    $CF::colon = '';
    $CF::slash = '/';
    $CF::use_hashes_for_cache_filenames = 0;

  } elsif (&Scoop::MyOS eq 'Win32') {
    $CF::colon = '';
    $CF::slash = '\\';
    $CF::use_hashes_for_cache_filenames = 0;

  } elsif (&Scoop::MyOS eq 'Mac') {
    $CF::slash = ':'; $CF::colon = ':';
    # because of the Mac's 32-char filename limitation, we need to include
    # a hash of the URL in cache filenames to avoid clashes. This may be
    # handy for other OSes too, but leave it Mac-only for now.
    $CF::use_hashes_for_cache_filenames = 1;
  }

  $CF::sitescooperdir = $FindBin::Bin;
  $Scoop::cwd = getcwd;

  &Scoop::verbose ("Sitescooper version ".$Scoop::VERSION
				  .", Copyright (c) 1999-2000 Justin Mason\n"
	  ."Sitescooper comes with ABSOLUTELY NO WARRANTY; for details\n"
	  ."see http://sitescooper.org/doc/gpl.html .\n");

}

sub Init {
  my $self = shift;

  $Scoop::cgimode = 0;			# for possible future CGI mode

  # TODO -- change all CF::s to member variables

  $CF::refresh = 0;
  $CF::debug = 0;
  $CF::debugdiffs = 0;			# set to 1 to break after diffing

  $CF::testsupportfile = undef;		# set when functional testing

  #$CF::just_caching = 0;
  $CF::cached_front_page_lifetime = 60;	# in minutes
  $CF::dump = 0;
  $CF::dumpprc = 0;

  $CF::verbose = 1;
  $CF::nowrite = 0;
  $CF::badcache = 0;
  $CF::grepmode = 0;
  undef $CF::prcdir;
  undef $CF::pilotinstdir;
  undef $CF::pilotinstapp;

  @{$self->{profiles}} = ( );
  @{$self->{profiles_cmdline}} = ( );

  # Text is the default on MacOS.
  if (&Scoop::MyOS eq 'Mac') {
    $self->{output_style} = $Scoop::OUT_TEXT;
    $CF::outputfilter = '__cat__';
    $CF::outputextn = 'txt';
    $CF::include_isilo_bookmarks = 0;
    $CF::use_convert_tool = 0;
    $CF::print_full_links = 0; # print full links

  } else {
    $self->{output_style} = $Scoop::OUT_HTML;
    $CF::outputfilter = 'isilo';
    $CF::outputextn = 'prc';
    $CF::include_isilo_bookmarks = 1;
    $CF::use_convert_tool = 1;
    $CF::print_full_links = 1; # do not print full links in output file for iSilo
  }
  $CF::converter_args = '';

  $CF::fileperpage = 0;
  $CF::nolinkrewrite = 0;
  $CF::defaultfilesizelimit = 300;		# in Kb
  $CF::filesizelimit = undef;
  $CF::storylimit = 0;
  $CF::linkslimit = 0;

  $CF::writeheader = 1;
  $CF::writefooter = 1;
  $CF::allowimgs = 1;

  $CF::use_only_cache = 0;
#CGI $Scoop::cgi = undef;
  $CF::admincmd = undef;

  @CF::sites_grep = ();
  @CF::site_files_to_read = ();
  @CF::site_choices = ();
  @CF::layout_site_files = ();
  @CF::cmdline_urls = ();
  %CF::caches_to_rename = ();
  $CF::keep_tmps = 0;
  $CF::disconnect_after_scoop = 0;
  $CF::preload_method = 'lwp';

  $Scoop::argv_settings = ();

  $Scoop::add_closing_tags = 1;		# close tags when cleaning HTML
  $Scoop::strip_empty_tag_sets = 0; 	# strip empty tag sets (broken atm)

  $CF::filename_template = "YYYY_MM_DD_Site";
  $CF::prc_title = "YYYY-Mon-DD: Site";

  # note the iSilo bookmark code here BTW.
  $self->{default_output_template} = q{

<htmlmainpage><html>
<head><title>__SITE_TITLE__</title>
__ISILO_BOOKMARK_LINKS__
</head>
<body bgcolor=white><h1>__SITE_TITLE__</h1>

__MAIN_BODY__

<p><hr><font size=1><i>
(End of snarf - copyright retained by original providers.
__SITE_RIGHTS__
Downloaded and converted by sitescooper; see __SITESCOOPER_HOME_URL__ )
</i></font>
</body></html>
</htmlmainpage>

<htmlsubpage><html>
<head><title>__SITE_TITLE__</title></head>
<body bgcolor=white>

__SUB_BODY__

</body></html>
</htmlsubpage>

<htmlstory>
<hr><font size=1><i>__SITE_NAME__:
<a HREF_EXTERNAL="__STORY_URL__">__SHORT_URL__</a></i></font><br>

__STORY_ANCHOR__ __ISILO_HEADLINE_ANCHOR__
    [<a href="__STORY_PREV_LINK__">&lt;&lt;</a>]
    [<a href="__STORY_UP_LINK__">^^</a>]
    [<a href="__STORY_NEXT_LINK__">&gt;&gt;</a>]<br>

__STORY_TEXT__
</htmlstory>
<textmainpage>__SITE_TITLE__


__MAIN_BODY__


(End of snarf - copyright retained by original providers.
__SITE_RIGHTS__
Downloaded and converted by sitescooper; see __SITESCOOPER_HOME_URL__ )
</textmainpage>

<textsubpage>__SUB_BODY__</textsubpage>

<textstory>------------
__SITE_NAME__: __STORY_URL__

__BOOKMARK_CHAR__ __HEADLINE__
__STORY_TEXT__
</textstory>

};
  $self->{output_template} = undef;

  $Scoop::conf = ();

# --------------------------------------------------------------------------

#if (defined $ENV{'REQUEST_METHOD'}) {
    # we're running from a CGI script, use CGI mode
    #$Scoop::cgimode = 1;
    #$Scoop::cgi = new CGI;
#}

# This is the placeholder for development debug flags.
# Add debugging stuff here, tagged with J M D (without the spaces ;).


# --------------------------------------------------------------------------
}

sub AddCmdlineSiteParam {
  my $self = shift;
  push (@Scoop::argv_settings, shift);
}

# --------------------------------------------------------------------------

sub ParseArgs {
  my ($self, @argv) = @_;
  local ($_);
  my @sites_grep = ();

  if ($Scoop::cgimode == 0) {
    my @macargs = split (' ', $Scoop::MAC_ARGS);
    unshift (@argv, @macargs);
    $Scoop::MAC_ARGS = '';

    while ($#argv >= 0) {
      $_ = shift @argv;

      if (/^-debug$/) {
	$CF::debug++;

      } elsif (/^-quiet$/) {
	$CF::verbose = 0;

      } elsif (/^-refresh/) {
	$CF::cached_front_page_lifetime = 0;
	$CF::refresh = 1;

      } elsif (/^-dump/) {
	$CF::dump = 1;
	$self->{output_style} = $Scoop::OUT_TEXT;
	$CF::use_convert_tool = 0;
	$CF::allowimgs = 0;

      } elsif (/^-dumpprc/) {
	$CF::dumpprc = 1;
	$CF::use_convert_tool = 1;

      } elsif (/^-doc/) {
	$self->{output_style} = $Scoop::OUT_DOC;
	$CF::fileperpage = 0;
	$CF::outputfilter = 'makedoc';
	$CF::include_isilo_bookmarks = 0;
	$CF::outputextn = 'prc';
	$CF::use_convert_tool = 1;
	$CF::allowimgs = 0;

      } elsif (/^-plucker/) {
        $self->{output_style} = $Scoop::OUT_HTML;
        $CF::fileperpage = 0;
        $CF::outputfilter = 'plucker';
	$CF::include_isilo_bookmarks = 0;
        $CF::outputextn = 'pdb';
        $CF::use_convert_tool = 1;
        $CF::allowimgs = 1;

      } elsif (/^-mplucker/) {
        $self->{output_style} = $Scoop::OUT_HTML;
        $CF::fileperpage = 1;
        $CF::outputfilter = 'plucker';
	$CF::include_isilo_bookmarks = 0;
        $CF::outputextn = 'pdb';
        $CF::use_convert_tool = 1;
        $CF::allowimgs = 1;

      } elsif (/^-isilo/) {
	$self->{output_style} = $Scoop::OUT_HTML;
	$CF::fileperpage = 0;
	$CF::outputfilter = 'isilo';
	$CF::include_isilo_bookmarks = 1;
	$CF::outputextn = 'prc';
	$CF::use_convert_tool = 1;
	$CF::allowimgs = 1;

      } elsif (/^-misilo/) {
	$self->{output_style} = $Scoop::OUT_HTML;
	$CF::fileperpage = 1;
	$CF::outputfilter = 'isilo';
	$CF::include_isilo_bookmarks = 1;
	$CF::outputextn = 'prc';
	$CF::use_convert_tool = 1;
	$CF::allowimgs = 1;

      } elsif (/^-richreader/) {
	$self->{output_style} = $Scoop::OUT_HTML;
	$CF::fileperpage = 0;
	$CF::outputfilter = 'richreader';
	$CF::include_isilo_bookmarks = 0;
	$CF::outputextn = 'pdb';
	$CF::use_convert_tool = 1;
	$CF::allowimgs = 1;		# TODO - not sure

      } elsif (/^-text/) {
	$self->{output_style} = $Scoop::OUT_TEXT;
	$CF::fileperpage = 0;
	$CF::outputfilter = '__path__';
	$CF::include_isilo_bookmarks = 0;
	$CF::outputextn = 'txt';
	$CF::use_convert_tool = 1;
	$CF::allowimgs = 0;

      } elsif (/^-html/) {
	$self->{output_style} = $Scoop::OUT_HTML;
	$CF::fileperpage = 0;
	$CF::outputfilter = '__path__';
	$CF::include_isilo_bookmarks = 0;
	$CF::outputextn = 'html';
	$CF::use_convert_tool = 1;
	$CF::allowimgs = 1;

      } elsif (/^-mhtml/) {
	$self->{output_style} = $Scoop::OUT_HTML;
	$CF::fileperpage = 1;
	$CF::outputfilter = '__path__';
	$CF::include_isilo_bookmarks = 0;
	$CF::outputextn = 'html';
	$CF::use_convert_tool = 1;
	$CF::allowimgs = 1;

      } elsif (/^-pipe/) {
	my $fmt = shift @argv;
	my $cmd = shift @argv;

	if ($fmt eq 'text') {
	  $self->{output_style} = $Scoop::OUT_TEXT;
	  $CF::fileperpage = 0;

	} elsif ($fmt eq 'html') {
	  $self->{output_style} = $Scoop::OUT_HTML;
	  $CF::fileperpage = 0;

	} elsif ($fmt eq 'mhtml') {
	  $self->{output_style} = $Scoop::OUT_HTML;
	  $CF::fileperpage = 1;
	} else {
	  &main::usage;
	}
	$CF::outputfilter = 'cmd: '.$cmd;
	$CF::outputextn = 'prc';
	$CF::use_convert_tool = 1;
	$CF::allowimgs = 1;		# probably ;)

      } elsif (/^-cvtargs/) {
	$CF::converter_args = shift @argv;

      } elsif (/^-admin$/) {
	$CF::admincmd = shift @argv;
	if ($CF::admincmd eq 'import-cookies')
				  { $Scoop::importcookies = shift @argv; }
	if ($CF::admincmd eq 'func-test')
				  { $CF::testsupportfile = shift @argv; }
	if ($CF::admincmd eq 'include-isilo-bookmarks')
				  { $CF::include_isilo_bookmarks = 1; }

      } elsif (/^-nolinkrewrite/) {
	$CF::nolinkrewrite = 1;
      } elsif (/^-fromcache/) {
	$CF::use_only_cache = 1;
      } elsif (/^-preload/) {
	$CF::preload_method = shift(@argv);
      } elsif (/^-limit/) {
	$_ = shift(@argv); s/\s*k\s*//i;
	$CF::filesizelimit = $_+0;
      } elsif (/^-maxlinks/) {
	$CF::storylimit = shift(@argv)+0;
      } elsif (/^-maxstories/) {
	$CF::linkslimit = shift(@argv)+0;
      } elsif (/^-nodates/) {
	$CF::filename_template = 'Site';
	$CF::prc_title = 'Site';
      } elsif (/^-nowrite/) {
	$CF::nowrite = 1;
      } elsif (/^-badcache/) {
	$CF::badcache = 1;
      } elsif (/^-grep/) {
	$CF::grepmode = 1;
      } elsif (/^-config/) {
	$CF::config = shift @argv;
      } elsif (/^-install/) {
	$CF::pilotinstdir = shift @argv;
      } elsif (/^-instapp/) {
	$CF::pilotinstapp = shift @argv;
      } elsif (/^-site$/) {
	push (@sites_grep, shift @argv);
      } elsif (/^-sites/) {
	push (@sites_grep, @argv); @argv = (); last;
      } elsif (/^-profile$/) {
	push (@{$self->{profiles_cmdline}}, shift @argv);
      } elsif (/^-profiles/) {
	push (@{$self->{profiles_cmdline}}, @argv); @argv = (); last;
      } elsif (/^-name/) {
	$Scoop::argv_name = shift (@argv);
      } elsif (/^-levels/) {
	$self->AddCmdlineSiteParam ("Levels: ". shift @argv);
      } elsif (/^-storyurl/) {
	$self->AddCmdlineSiteParam ("StoryURL: ". shift @argv);
      } elsif (/^-set/) {
	$self->AddCmdlineSiteParam (shift (@argv) .": ". shift @argv);
      } elsif (/^-keep-tmps/) {
	$CF::keep_tmps = 1;
      } elsif (/^-disc/) {
	$CF::disconnect_after_scoop = 1;

      } elsif (/^-noheaders/) {
	$CF::writeheader = 0;
      } elsif (/^-nofooters/) {
	$CF::writefooter = 0;
      } elsif (/^-outputtemplate/) {
	$_ = shift @argv;
	open (IN, "< $_") or die "cannot open output template: $_\n";
	$self->{output_template} = join ("\n", <IN>);
	close IN;

      } elsif (/^-filename/) {
	$CF::filename_template = shift @argv;
      } elsif (/^-prctitle/) {
	$CF::prc_title = shift @argv;

      } elsif (/^-stdout-to/) {
	$_ = shift @argv; close (STDOUT);
	open (STDOUT, ">> ".$_) or die "failed to redirect STDOUT to $_\n";

      } elsif (/^-/) {
	&main::usage;

      } elsif (/\.scp$/i) {
	$CF::config = $_;

      } else {
	unshift @argv, $_; last;
      }
    }

    @CF::cmdline_urls = @argv;
    $Scoop::userid = $<;

  } else {
    # load some things from CGI parameters
    #CGI@CF::cmdline_urls = ($Scoop::cgi->param ('url'));

    #CGI@CF::sites_grep = $Scoop::cgi->param ('sites');

    #CGI$CF::debug = $Scoop::cgi->param ('debug');
    #CGI$self->{output_style} = $Scoop::cgi->param ('output_style');
    #CGI$CF::nowrite = $Scoop::cgi->param ('nowrite');
    #CGI$CF::refresh = $Scoop::cgi->param ('refresh');
    #CGI$Scoop::userid = $Scoop::cgi->param ('userid');
    #CGI&ScoopCGI::get_cookie;
    #CGI $Scoop::password = $Scoop::cgi->param ('password');
    # REVISIT -- use a cookie to store userid and password

    #CGI$Scoop::pilotinstdir = undef;
  }

  if (!$Scoop::cgimode) {
    my $key;
    foreach $key (@sites_grep) {
      if (!-r $key) {
	warn "Failed to read -site argument \"$key\".\n";
	next;
      }

      # it's a site file.
      push (@CF::site_files_to_read, $key);
      push (@CF::sites_grep, $key);
      # TODO: add support for -sitesmatching pattern or similar...
    }
  }
}

# --------------------------------------------------------------------------

sub CreateConfigIfNeeded {
  my $self = shift;

  my $default_config = $CF::sitescooperdir.$CF::slash."sitescooper.cf";
  if (-r "/etc/sitescooper.cf") { $default_config = "/etc/sitescooper.cf"; }

  # UNIX platforms: use ~/.sitescooper/sitescooper.cf if it exists.
  # Otherwise use /etc/sitescooper.cf if that exists.
  # All platforms use {sitescooperdir}/sitescooper.cf if that exists.
  #
  if (&Scoop::MyOS eq 'UNIX') {
    my $homecf = $ENV{'HOME'}."/.sitescooper/sitescooper.cf";;
    if (!defined $CF::config && -r $homecf) { $CF::config ||= $homecf; }
  }

  # UNIX platforms: create a default configuration file. This makes it
  # easier to support rpm installation.
  #
  $CF::config ||= $default_config;
  if (!-r $CF::config) {
    die "Cannot find default sitescooper.cf file, ".
    	"please use the '-config' argument!\n";
  }

  if (&Scoop::MyOS eq 'UNIX') {
    $CF::tmpdir ||= $ENV{'HOME'}."/.sitescooper";
    my $homecf = $ENV{'HOME'}."/.sitescooper/sitescooper.cf";;

    if (!-r $homecf) {
      (-d $CF::tmpdir) or
	  mkdir ($CF::tmpdir, 0777) or die "failed to mkdir '$CF::tmpdir'\n";

      warn "Copying default config to \"$homecf\".\n".
	"Edit this if you need to change any configuration settings.\n\n";

      copy ($CF::config, $homecf) or die "cannot copy $CF::config to $homecf\n";
      $CF::config = $homecf;
    }
  }

  if (!defined $CF::config || !-r $CF::config) {
    my $cfg = $CF::config; $cfg ||= "(unset)";
    die ("Cannot find configuration file \"$cfg\"; ".
    	"please use the '-config' argument!\n");
  }
}

sub ReadConfig {
  my $self = shift;

  $self->CreateConfigIfNeeded();

  $self->ReadRawConfig();
  $self->SetupDefaultConfig();
  $self->ParseConfig (\@Scoop::conf, \@Scoop::conflines);
  undef @Scoop::conf;		# save memory
  undef @Scoop::conflines;

  $self->InterpretBasicConfig();
  $self->LoadAllModules();
}

# --------------------------------------------------------------------------

sub ReadRawConfig {
  my $self = shift;

  @Scoop::conf = ();
  @Scoop::conflines = ();
  &Scoop::verbose ("Reading configuration from \"$CF::config\".");
  {
    open (IN, "<$CF::config") || die "cannot read $CF::config\n";
    @Scoop::conf = (<IN>); close IN;

    my $i;
    for ($i=0; $i<$#Scoop::conf+1; $i++) {
      push (@Scoop::conflines, $CF::config.":".($i+1));
    }
  }

  if ($CF::debugdiffs) {
    &Scoop::dbg ("debugging, will exit after diff");
  }
}

# --------------------------------------------------------------------------

sub SetupDefaultConfig {
  my $self = shift;

  $CF::outdir = '';
  $CF::cached_front_page_lifetime /= (24*60);	# convert to days
  $CF::expiry_days = 7.0;
  $CF::sharedcache = undef;

  %AllSCFs::sites = ();

  @AllSCFs::layout_order = ();
  %AllSCFs::layouts = ();
  $AllSCFs::have_layouts = 0;

  @AllSCFs::exception_order = ();
  %AllSCFs::exceptions = ();
  $AllSCFs::have_exceptions = 0;

  undef $CF::tmpdir;
  if (&Scoop::MyOS eq 'UNIX') {
    $CF::tmpdir = $ENV{'HOME'}."/.sitescooper";
  }
  $CF::tmpdir ||= $ENV{'TMPDIR'};
  $CF::tmpdir ||= $ENV{'TEMP'};
  if (!defined $CF::tmpdir && &Scoop::MyOS eq 'Win32' && defined $ENV{'WINDIR'})
  {
    $CF::tmpdir = $ENV{'WINDIR'}."\\Temp";
  }

  $CF::diff = 'diff';
  if (&Scoop::MyOS eq 'Win32') { $CF::diff = "diff.exe"; }
  if (&Scoop::MyOS eq 'Mac') { $CF::diff = ""; }	# use Algorithm::Diff
  $CF::checked_for_diff = 0;

  $CF::makedoc = 'makedoc';
  if (&Scoop::MyOS eq 'Win32') { $CF::makedoc = "makedocw.exe"; }

  $CF::plucker = 'plucker-build';
  $CF::pluckargs = '';

  $CF::isilo = 'iSilo386'; $CF::isiloargs = '-y -U';
  $CF::isilomultipageargs = '-d9 -y -U';
  if ($CF::allowimgs) {
    $CF::isiloargs .= ' -y -U -Is__IMAGE_MAX_WIDTH__ -Ic -Id';
  }

  if (&Scoop::MyOS eq 'Win32') { $CF::isilo = "iSiloC32.exe"; }

  $CF::richreader = 'HTML2Doc'; $CF::richargs = '';
  if (&Scoop::MyOS eq 'Win32')
	  { $CF::richreader = "HTML2Doc.exe"; $CF::richargs = ''; }
  # Note that currently there is no HTML2Doc for UNIX platforms; it's
  # supported here anyway for future-proofing.

  &Scoop::set_got_intr_behaviour ('exit');
  $SIG{'INT'} = \&Scoop::got_intr;
  $SIG{'TERM'} = \&Scoop::got_intr;

  $CF::proxyhost = undef;
  $CF::proxyport = 80;
}

# ---------------------------------------------------------------------------

sub InterpretBasicConfig {
  my $self = shift;

  if (!defined $CF::tmpdir) {
    warn "Warning: cannot work out TmpDir, please set it manually\n".
	  "in the configuration section of the script.\n";
  }
  if (!defined $CF::sitescooperdir) {
    warn "Warning: cannot work out SitescooperDir, please set it manually\n".
	  "in the configuration section of the script.\n";
  }
  $self->make_basic_dirs;	# make tmpdir and user_tmpdir first off

# ---------------------------------------------------------------------------

  if (!defined $CF::pilotinstdir && !$Scoop::cgimode)
  {
    if (&Scoop::MyOS eq 'Win32' && !defined $CF::pilotinstapp) {
      $CF::pilotinstapp = "***USE_MODULE***";

    } elsif (&Scoop::MyOS eq 'UNIX') {
      my $pilot_mgr_dir = $ENV{'HOME'}."/.pilotmgr/Installer";
      my $jpilot_file = $ENV{'HOME'}."/.jpilot/jpilot_to_install";

      if (defined $CF::pilotinstapp) {
	# see if one of the built-in support for UNIX pilot desktops is
	# being used.
	if ($CF::pilotinstapp =~ /pilot.*manager/i)
	  { $CF::pilotinstdir = $pilot_mgr_dir; undef $CF::pilotinstapp; }
	elsif ($CF::pilotinstapp =~ /gnome.*pilot/i)
	  { $CF::pilotinstapp = "gpilot-install-file --later"; }
	elsif ($CF::pilotinstapp =~ /jpilot/i)
	  { $CF::pilotinstapp = "***ADD_TO_MANIFEST*** ".$jpilot_file; }

      } else {
	$CF::pilotinstapp = "***USE_MODULE***";		# use the module
      }
    }
  }

# ---------------------------------------------------------------------------

  # we may not have read the sites directory by this stage.
  if (!defined $CF::sitesdir) {
    &Scoop::dbg ("SitesDir was not specified, trying to guess it...");

    my $possible = "sites";
    if (!defined $CF::sitesdir && -d $possible) { $CF::sitesdir = $possible; }

    if (&Scoop::MyOS eq 'UNIX') {
      $possible = $ENV{'HOME'}."/sites";
      if (!defined $CF::sitesdir && -d $possible) { $CF::sitesdir = $possible; }

      # default for UNIX platforms
      $CF::sitesdir ||= $CF::user_tmpdir.$CF::slash."sites";

    } else {
      # default for non-UNIX platforms
      $CF::sitesdir ||= $CF::sitescooperdir.$CF::slash."sites";
    }
  }
}

sub LoadAllModules {
  my $self = shift;

  # ensure we can find our external modules. Also ensure they override any
  # others already in the path, as some distros and versions of perl include
  # old versions of LWP etc.
  #
  #unshift (@INC, $CF::sitescooperdir.$CF::slash."lib");
  #unshift (@INC, $CF::sitescooperdir.$CF::slash."site_perl");

  require LWP::UserAgent;
  require URI::URL;
  require HTTP::Date;
  require HTTP::Cookies;
  require HTTP::Request::Common;
  require HTML::Entities;
  require HTML::Parser;
  require HTML::Filter;

  require Sitescooper::Robot;
  require Sitescooper::SCF;
  require Sitescooper::StripTablesFilter;
  require Sitescooper::UserAgent;
  require Sitescooper::ConsolePasswordAsker;
  require Sitescooper::LWPHTTPClient;
  require Sitescooper::ForkHTTPClient;
  require Sitescooper::PreloadURLProcessor;
  require Sitescooper::NewsHound;

  require PDA::PilotInstall;

# ---------------------------------------------------------------------------

  $Scoop::useragent = new Sitescooper::UserAgent;
  $Scoop::useragent->env_proxy;
  $Scoop::useragent->agent ("sitescooper/$Scoop::VERSION ($Scoop::home_url) ".
		  $Scoop::useragent->agent);
  $Scoop::useragent->max_size (1024*1024*2);	# 2-meg file limit

  $Scoop::cookie_jar = HTTP::Cookies::Netscape->new;
  $Scoop::pua = undef;

  my $asker = new Sitescooper::ConsolePasswordAsker;
  $Scoop::useragent->set_password_asker ($asker);

  if ($CF::preload_method eq 'lwp') {
    eval q{
      $self->{httpclient} = new Sitescooper::LWPHTTPClient ($self);
      1;
    } or die "eval for preload class failed";
  } elsif ($CF::preload_method =~ /^fork(\d*)/) {
    my $num = $1; ($num eq '') and $num = 4;
    eval q{
      $self->{httpclient} = new Sitescooper::ForkHTTPClient ($self, $num);
      1;
    } or die "eval for preload class failed";
  } else {
    warn "Unknown preload method '".$CF::preload_method."'.\n";
    &main::usage();
  }

  $self->{httpclient}->init();

  if (!defined $CF::pilotinstdir && !$Scoop::cgimode
    	&& !defined $CF::pilotinstapp)
  {
    # write PRCs to cwd if nothing is set.
    &Scoop::verbose ("Warning: since no PilotInstallDir was specified".
    " in the configuration,\nPRC files will be saved to current directory.\n");
    $CF::pilotinstdir = $Scoop::cwd;

  } elsif (defined $CF::pilotinstdir) {
    # just write them direct to the install dir.
    $CF::prcdir = $CF::pilotinstdir;
  }
}

# ---------------------------------------------------------------------------

sub set_got_intr_behaviour {
  $Scoop::got_intr_behaviour = shift;
  $Scoop::got_intr_flag = 0;
}

sub got_intr {
  my $signame = shift;
  (&Scoop::MyOS eq 'UNIX') and system ("stty echo");

  if ($Scoop::got_intr_behaviour eq 'exit') {
    die "got signal SIG$signame, exiting.\n";
  } else {
    die "got signal SIG$signame, skipping site...\n";
    $Scoop::got_intr_flag = 1;
  }
}

# ---------------------------------------------------------------------------

sub ReadSites {
  my $self = shift;
  @Scoop::conf = ();
  @Scoop::conflines = ();

  $self->ReadSiteChoices();
  $self->ScanSitesDir();
  $self->ReadSiteFiles();
  $self->ReadNewsHoundProfiles();

  # parse the site files. We've already read the config file BTW.
  $self->ParseConfig (\@Scoop::conf, \@Scoop::conflines);
  undef @Scoop::conf;		# save memory
  undef @Scoop::conflines;

  $self->PolishConfig;

  if (&Scoop::MyOS eq 'Mac') {
    if (defined $CF::cmdline_urls[0] && $CF::cmdline_urls[0] eq $CF::sitescooperdir) {
      # REVISIT -- I do not know why this happens :(
      shift (@CF::cmdline_urls);
    }
  }

# ---------------------------------------------------------------------------

  if ($#CF::cmdline_urls > -1) {
    %AllSCFs::sites = ();

    my ($url, $confline, @conf, @conflines, $i);
    foreach $url (@CF::cmdline_urls)
    {
      # if it's a local file URL, switch around the slashes (for windows)
      if (&Scoop::MyOS eq 'Win32' && $url =~ m,file:///,i) {
	$url =~ s/\\/\//g;
      }
      # REVISIT -- I don't know what to do in the same case for MacOS ;)

      if (-r $url) {
	if ($url =~ m,^/,) {
	  $url = 'file://'.$url;
	} else {
	  $url = "file://".$Scoop::cwd."/".$url;
	}
      }

      if ($url =~ m,(http|file)://[^/]+$,i) { $url .= '/'; }
      push (@conf, "URL: ".$url);

      if (defined $Scoop::argv_name) {
	$_ = $Scoop::argv_name;
      } else {
	if ($url =~ m,/([^/]+)$,) {
	  $_ = $1;
	  if (length ($_) > 40) {
	    # trim out spare stuff to keep it short.
	    s,^([^:]+://[^/]+)/.*/([^/]+$),$1/.../$2,i; #/
	  }
	} else {
	  $_ = $url;
	}
      }
      push (@conf, "Name: ".$_);
      push (@conf, "StoryLifetime: 0");	# never used cached stuff
      push (@conf, @Scoop::argv_settings);

      for ($i=0; $i<$#conf+1; $i++) { push (@conflines, $url.":0"); }
      $self->ParseConfig (\@conf, \@conflines);
    }
  }

# ---------------------------------------------------------------------------

  {
    my ($mday, $mon, $year, $monstr) = &get_date;
    my ($min, $hr, $wdaystr) = &get_extra_date;

    $CF::filename_template =~ s/YYYY/ sprintf ("%04d", $year); /eg;
    $CF::filename_template =~ s/MM/ sprintf ("%02d", $mon); /eg;
    $CF::filename_template =~ s/Mon/ sprintf ("%3s", $monstr); /eg;
    $CF::filename_template =~ s/DD/ sprintf ("%02d", $mday); /eg;
    $CF::filename_template =~ s/Day/ sprintf ("%3s", $wdaystr); /eg;
    $CF::filename_template =~ s/hh/ sprintf ("%02d", $hr); /eg;
    $CF::filename_template =~ s/mm/ sprintf ("%02d", $min); /eg;

    $CF::prc_title =~ s/YYYY/ sprintf ("%04d", $year); /eg;
    $CF::prc_title =~ s/MM/ sprintf ("%02d", $mon); /eg;
    $CF::prc_title =~ s/Mon/ sprintf ("%3s", $monstr); /eg;
    $CF::prc_title =~ s/DD/ sprintf ("%02d", $mday); /eg;
    $CF::prc_title =~ s/Day/ sprintf ("%3s", $wdaystr); /eg;
    $CF::prc_title =~ s/hh/ sprintf ("%02d", $hr); /eg;
    $CF::prc_title =~ s/mm/ sprintf ("%02d", $min); /eg;
  }

  %Scoop::already_seen = ();
  %Scoop::last_modtime = ();
  %Scoop::oldest_already_seen = ();
  @Scoop::seen_this_time = ();
  $Scoop::failed_to_cvt = 0;

  if (defined $CF::testsupportfile) {
    &read_test_support ($CF::testsupportfile);
  }

  $self->make_dirs;		# Note: this will chdir
  $self->generate_output_filenames (keys %AllSCFs::sites);
  $self->handle_admin_cmds();

  if ($CF::use_convert_tool) {
    if (defined $CF::prcdir && !-d $CF::prcdir) {
      mkdir ($CF::prcdir, 0755) || die "failed to mkdir '$CF::prcdir'\n";
    }
    if (defined $CF::pilotinstdir && !-d $CF::pilotinstdir) {
      mkdir ($CF::pilotinstdir, 0755) || die "failed to mkdir '$CF::pilotinstdir'\n";
    }
  }

  $self->expire_old_cache_files;
  $Scoop::useragent->load_logins();
  $self->read_state;

  if (defined $CF::pilotinstapp && $CF::pilotinstapp eq '***USE_MODULE***')
  {
    $CF::installer = new PDA::PilotInstall;
    $CF::installer->read_config_file ($CF::user_tmpdir.$CF::slash."inst.txt");
  }
}

# ---------------------------------------------------------------------------

sub read_a_profile {
  my ($self, $file) = @_;

  Scoop::verbose ("Reading profile from file \"$file\"...");

  my $profile = Sitescooper::NewsHound::LoadProfile($file);
  if ($profile == 0) {
    warn "Can't load profile '$file': $!\n";
  } else {
    push (@{$self->{profiles}}, $profile);
  }
}

sub ReadNewsHoundProfiles {
  my $self = shift;
  my $file;

  my $profdir = $self->{profilesdir};
  if (defined $profdir) {
    foreach $file (<$profdir/*.nhp>) {
      next if (-d $file);             # skip directories
      $self->read_a_profile ($file);
    }
  }

  foreach $file (@{$self->{profiles_cmdline}}) {
    if (!-r $file) {
      warn "Failed to read -profile argument \"$file\".\n";
      next;
    }
    $self->read_a_profile ($file)
  }
}

# ---------------------------------------------------------------------------

sub handle_admin_cmds {
  my $self = shift;
  if (defined $CF::admincmd) {
    if ($CF::admincmd eq 'dump-sites') {
      my ($robot, $outdir);

      foreach $robot (@Scoop::robots) {
        my $outdir = $robot->{outdir};
	my $url = $robot->{url};
	my $title = $robot->{sitename};
	$title =~ s,\t, ,g; $title =~ s,^\d+-\S+-\d+: ,,g;

	my $base = $robot->{outtmp};
	$base =~ s,^.*${CF::slash}(\S+?)\.tmp$,$1,o;

	my $scf = $AllSCFs::sites{$url};
	my $site = $scf->{site_defined_at};

	$site =~ s/:\d+$//; $site =~ s/^.*${CF::slash}(\S+?)$/$1/o;

	# foobar.site	http://www.foobar.com/	Foo Bar	1999_01_01_Foo_Bar
	print "$site\t$url\t$title\t$base\n";
      }
      exit;

    } elsif ($CF::admincmd eq 'journal') {
      open (JOURNAL, "> ".$CF::tmpdir.$CF::slash."journal")
	  or die "cannot write to ".$CF::tmpdir.$CF::slash."journal!\n";

    } elsif ($CF::admincmd eq 'import-cookies') {
      if (!defined $Scoop::importcookies) { main::usage(); }
      if (!-r $Scoop::importcookies) {
	die "cannot open \"$Scoop::importcookies\"!\n";
      }
      warn "Importing Netscape-format cookie jar from ".
      		"\"$Scoop::importcookies\"...\n";
      $Scoop::cookie_jar->load ($Scoop::importcookies);
      warn "Cookie jar now looks like:\n".$Scoop::cookie_jar->as_string;
      %AllSCFs::sites = ();
      @CF::cmdline_urls = ();
      @Scoop::robots = ();
      # and carry on to exit.

    } elsif ($CF::admincmd eq 'func-test') {
      # do nothing right now, it's handled later
    } elsif ($CF::admincmd eq 'include-isilo-bookmarks') {
      # do nothing here

    } else { &main::usage; }
  }
}

# ---------------------------------------------------------------------------

sub Run {
  my $self = shift;
  # to do all the conversions at the end:
  #$self->get_all_sites;
  #foreach $robot (@Scoop::robots) {
  #  $self->convert_output ($robot);
  #}
  #(TODO: this hasn't been modified to use the scf object.)

  # to do them as each site is scooped:
  $self->get_all_sites (1);
}

# ---------------------------------------------------------------------------

sub Finish {
  my $self = shift;
  delete $self->{httpclient};
  $self->write_state;
  $Scoop::useragent->save_logins();
  $Scoop::cookie_jar->save ($CF::user_tmpdir.$CF::slash."cookies");

  if (defined $CF::pilotinstapp && $CF::pilotinstapp eq '***USE_MODULE***')
  {
    $CF::installer->write_config_file ($CF::user_tmpdir.$CF::slash."inst.txt");
  }

  if ($CF::disconnect_after_scoop) {
    $self->disconnect;
  }

  &Scoop::verbose ("Finished!");
}

# ---------------------------------------------------------------------------

sub ParseConfig (\@\@) {
  my $self = shift;
  local ($_);
  my ($conf, $conflines) = @_;

  my $cf = undef;
  my $scf = undef;
  my $confline;

  foreach $_ (@$conf) {
    $confline = shift @$conflines;
    s/#.*$//; s/^\s+//; s/\s+$//g; next if (/^$/);
    if (!defined $confline) {
      &Scoop::dbg ("oops! confline not set for $_");
    }

    # process environment variable references: ${ENVVARNAME}
    s/\$\{(\S+?)\}/
	  defined($ENV{$1}) ? $ENV{$1} : "";
    /ge;
    s/\$HOME/$ENV{'HOME'}/ge;		# always supported

    if (defined $scf && $scf->is_in_scoped_statement()) {
      $scf->ParseConfigScopedLine ($self, $_, $confline) and next;
    }

    s/^(\S+:)\s+/$1 /;		# easier to read this way ;)
    study;

    # now handle the general config parameters
    /^ProxyHost: (.*)$/ and ($CF::proxyhost = $1), next;
    /^ProxyPort: (.*)$/ and ($CF::proxyport = $1+0), next;
    /^TmpDir: (.*)$/ and ($CF::tmpdir = $1), next;
    if (/^SitescooperDir: (.*)$/) {
      $CF::sitescooperdir = $1;
      if (!-d $CF::sitescooperdir) {
	warn "${CF::config}: SitescooperDir directory does not exist: $CF::sitescooperdir\n";
      }
      next;
    }
    if (/^SitesDir: (.*)$/) { $CF::sitesdir = $1; next; }
    if (/^ProfilesDir: (.*)$/) { $self->{profilesdir} = $1; next; }

    # provide support for reading command line flags from config file
    # for MacOS (has no inherent command line concept)
    if (/^CommandLine: (.*)$/) {
      $self->ParseArgs(split (' ',$1)); next;
    }

    /^MakeDoc: (.*)$/ and ($CF::makedoc = $1), next;
    /^iSilo: (.*)$/ and ($CF::isilo = $1), next;
    /^Plucker: (.*)$/ and ($CF::plucker = $1), next;
    /^HTML2Doc: (.*)$/ and ($CF::richreader = $1), next;
    /^Diff: (.*)$/ and ($CF::diff = $1), next;
    /^TextSaveDir: (.*)$/ and ($CF::outdir = $1), next;
    /^PilotInstallDir: (.*)$/ and ($CF::pilotinstdir = $1), next;
    /^PilotInstallApp: (.*)$/ and ($CF::pilotinstapp = $1), next;
    /^SharedCacheDir: (.*)$/ and ($CF::sharedcache = $1), next;
    /^ExpireCacheAfter: (.*)$/ and ($CF::expiry_days = $1+0.0), next;
    /^BookmarkChar: (.*)$/ and ($main::bookmark_char = eval(qq!"$1"!)), next;

    if (/^CachedPageLifetime: (.*)$/) {
      (!$CF::refresh) and
    		($CF::cached_front_page_lifetime = ($1+0) / (24*60));
      next;
    }

    # next the site file config params
    if (/^URL: (.*)$/) {
      my $url = &expand_url_magic ($1);
      if (!defined $url) {
	&Scoop::sitewarn_file_line ($confline, "Bad URL in site file: $_\n");
      }

      if ($url !~ m,^(http|file)://,i) { $url = 'http://'.$url; }
      if ($url =~ m,(http|file)://[^/]+$,i) { $url .= '/'; }

      $scf = new Sitescooper::SCF ();
      $scf->set_defaults ($url, $confline);
      $AllSCFs::sites{$url} = $scf;
      next;
    }

    # LayoutURL is similar to URL, but defines a layout for a specific
    # pattern. If an URL falls within this pattern, and parameters are
    # defined for this layout but not defined by the site file, the
    # layout parameters will be used.
    #
    if (/^LayoutURL: (.*)$/) {
      my $url = &expand_url_magic ($1);

      if ($url !~ m,^(http|file)://,i) { $url = 'http://'.$url; }
      if ($url =~ m,(http|file)://[^/]+$,i) { $url .= '/'; }

      # allow extra parameters to be added to an existing layout
      if (defined $AllSCFs::layouts{$url}) {
	$scf = $AllSCFs::layouts{$url};
      } else {
	$scf = new Sitescooper::SCF ();
	$scf->set_defaults ($url, $confline);
	$AllSCFs::layouts{$url} = $scf;
      }

      next;
    }

    # ExceptionURL is like LayoutURL, but it takes priority over
    # both LayoutURL and the normal site file rules. This way you
    # can define bits of a site that uses different layouts, caching
    # rules etc. by matching pages' URLs against the ExceptionURL
    # regular expression.
    #
    if (/^ExceptionURL: (.*)$/) {
      my $url = &expand_url_magic ($1);

      if ($url !~ m,^(http|file)://,i) { $url = 'http://'.$url; }
      if ($url =~ m,(http|file)://[^/]+$,i) { $url .= '/'; }

      # allow extra parameters to be added to an existing exception
      if (defined $AllSCFs::exceptions{$url}) {
	$scf = $AllSCFs::exceptions{$url};
      } else {
	$scf = new Sitescooper::SCF ();
	$scf->set_defaults ($url, $confline);
	$AllSCFs::exceptions{$url} = $scf;
      }

      next;
    }

    if (!defined $scf) {
      &Scoop::sitewarn_file_line ($confline,
	  "Configuration line invalid (needs URL line first?):\n  $_\n");
    }
    $scf->ParseConfigLine ($self, $_, $confline) and next;

    &Scoop::sitewarn_file_line ($confline, "Unrecognised:\n  $_\n");
  }

  if (defined $scf && $scf->is_in_scoped_statement()) {
    &Scoop::sitewarn_file_line ($confline,
	  "Fell off end of ".$scf->scoped_statement_type().
	  " statement!\n");
  }
}

# ---------------------------------------------------------------------------

sub PolishConfig {
  my $self = shift;
  if (defined $CF::proxyhost) {
    $Scoop::useragent->proxy
	  (['http', 'ftp'], "http://${CF::proxyhost}:${CF::proxyport}/");
  }

  @AllSCFs::layout_order = ();
  my $pat;
  foreach $pat (sort { length($a) <=> length($b) } keys %AllSCFs::layouts) {
    $AllSCFs::layouts{$pat}->{active} = 0; # not treated like sites
    push (@AllSCFs::layout_order, $pat);
    $AllSCFs::have_layouts = 1;
  }
  &Scoop::dbg ("site layouts defined: ", join(' ', @AllSCFs::layout_order));

  @AllSCFs::exception_order = ();
  foreach $pat (sort { length($a) <=> length($b) } keys %AllSCFs::exceptions)
  {
    $AllSCFs::exceptions{$pat}->{active} = 0; # not treated like sites
    push (@AllSCFs::exception_order, $pat);
    $AllSCFs::have_exceptions = 1;
  }
  &Scoop::dbg ("site exceptions defined: ",
  			join(' ', @AllSCFs::exception_order));

  # CRG: Added $CF::checked_for_diff to prevent diff check that follows
  # which would fail.
  if ($CF::diff eq 'MODULE') { $CF::diff = ''; $CF::checked_for_diff = 1; }

  if (&Scoop::MyOS eq 'Win32' && $CF::checked_for_diff == 0) {
    my $file = $CF::user_tmpdir.$CF::slash."tstdiff.txt";

    open (OUT, "> $file"); close OUT;
    warn "Checking for availability of the \"diff.exe\" command...\n";

    my $difffailed = 0;
    if (open (IN, $CF::diff." \"$file\" \"$file\" |")) {
      1 while (<IN>);
      close IN;
      $difffailed = ($? >> 8);

    } else {
      $difffailed = 1;
    }

    if ($difffailed) {
      warn "\n".
	"$CF::diff could not be found and run. Using perl module\n".
	"Algorithm::Diff instead.\n";

      $CF::diff = "";		# use module instead
    }
    unlink $file; $CF::checked_for_diff = 1;
  }

  if ($CF::diff eq '') {
    &Scoop::dbg ("using Algorithm::Diff module to diff pages.");
    eval q{ use Algorithm::Diff qw(diff); 1; }
		  or die "Cannot use built-in diff support, perl module\n".
			  "Algorithm::Diff not found: $@\n";
  }

  $self->parse_output_template();
}

# ---------------------------------------------------------------------------

sub ReadSiteChoices {
  my $self = shift;
  my $choicefile = $CF::tmpdir.$CF::slash."site_choices.txt";

  if (!-r $choicefile) {
    $self->EditSiteChoices ($choicefile);	# or create it in this case
  }

  &Scoop::verbose ("Using site choices from \"$choicefile\".");
  open (IN, "<$choicefile") or return;

  my $samplesdir = $CF::sitescooperdir.$CF::slash."site_samples";
  while (<IN>) {
    if (/^\s*\[\s*x\s*\]/i) {
      while (<IN>) {
	if (/^\s*Filename:\s*(\S+)\s*$/) {
	  $_ = $1; s/\[samples\]/${samplesdir}/g;
	  if (/layouts\.site/) {
	    push (@CF::layout_site_files, $_);
	  } else {
	    # &Scoop::dbg ("site choice: $_");
	    push (@CF::site_choices, $_);
	  }
	  last;
	}
      }
    }
  }
  close IN;
}

# ---------------------------------------------------------------------------

sub find_sites {
  return unless (/\.site?$/i && -f $_);
  push(@TmpGlobal::site_files_found, $File::Find::name);
}

sub ScanSitesDir {
  my $self = shift;
  my ($file, $key);
  my %sites_grep = ();
  my %read_sites = ();

  foreach $key (@CF::sites_grep) { $sites_grep{$key} = 1; }

  if ($#CF::sites_grep >= 0) {
    &Scoop::verbose ("Restricting to sites: ".join (' ', @CF::sites_grep));

  } elsif ($#CF::site_choices >= 0) {
    # only scoop sites from the site_choices list if the -site argument
    # was not used.
    foreach $key (@CF::site_choices) {
      &Scoop::dbg ("Adding site-choices site: ".$key);
      if (-r $key) {
	push (@CF::site_files_to_read, $key);
      } else {
	warn "Failed to read site-choices site file: $key\n";
      }
    }
  }

  foreach $key (@CF::layout_site_files) {
    &Scoop::dbg ("Using layout: $key");
    if (-r $key) { push (@CF::site_files_to_read, $key); }
  }

  if ($#CF::cmdline_urls >= 0) {
    # we're only snarfing the command-line URLs, skip the predefined sites
    return;
  }

  if (defined $CF::sitesdir && -d $CF::sitesdir) {
    @TmpGlobal::site_files_found = (); find(\&find_sites, $CF::sitesdir);

    foreach $file (@TmpGlobal::site_files_found) {
      next if ($file =~ /(\.swp$|core|\.bak$|\~$|^#)/);	# skip backups, etc.
      next if (-d $file);		# skip directories

      if ($#CF::sites_grep >= 0) {
	my $base = $file; $base =~ s,^.*[\/\\:]([^\/\\:]+)$,$1,g;
	&Scoop::dbg ("checking if site file is wanted: $file");
	next unless (defined $sites_grep{$base}
		|| defined $sites_grep{$file}
		|| $file =~ /layouts\.site/i);
      }
      push (@CF::site_files_to_read, $file);
    }
    undef @TmpGlobal::site_files_found;
  }
}

sub ReadSiteFiles {
  my $self = shift;
  if ($#CF::site_files_to_read < 0) {
    warn "\n".
      "No sites were read -- the site_choices.txt file is empty, or the\n".
      "\"sites\" directory could not be found.\n\n";
  }

  if ($#CF::cmdline_urls >= 0) {
    return;                            # don't read the real site files
  }

  my $file;
  my %read_sites = ();
  foreach $file (@CF::site_files_to_read) {
    next if (defined $read_sites{$file});
    $read_sites{$file} = 1;		# don't read the same file twice

    if (open (IN, "<$file")) {
      my $line = 0;
      while (<IN>) {
	push (@Scoop::conf, $_);
	push (@Scoop::conflines, "$file:$line"); $line++;
      }
      close IN;
      &Scoop::dbg ("Scooping site from file \"$file\".");

    } else {
      &Scoop::sitewarn_file_line ("$file:0", "Cannot read $file\n");
    }
  }
}

# ---------------------------------------------------------------------------

sub EditSiteChoices {
  my $self = shift;
  my $choicefile = shift;
  my $samplesdir = $CF::sitescooperdir.$CF::slash."site_samples";

  if (!-d $samplesdir) {
    warn "Cannot find 'site_samples' directory, not creating site_choices file.\n".
    	"(looked for '$samplesdir')\n";
    return;
  }

  if ($#CF::sites_grep >= 0) {
    warn "'-site' argument used, not creating site_choices file.\n";
    return;
  }

  warn "Creating/editing \"site_choices.txt\" file...\n";

  my %chosen_sites = ();
  foreach $_ (@CF::site_choices) { $chosen_sites{$_} = 1; }

  if (!open (CHOICE, ">$choicefile")) {
    warn "Cannot create \"site_choices.txt\" file $choicefile\n";
    return;
  }

  print CHOICE <<EOHDR;
Please pick the site files you wish to use here.  Put an X in the box
beside the sites you wish to scoop.

If you want to use the traditional 'sites' directory, or you have your own
site files not in this list, then do not put an X in any of the boxes.
Sitescooper will supplement what you have ticked here with the contents
of your 'sites' directory, if it exists.

EOHDR

  @TmpGlobal::site_files_found = (); find(\&find_sites, $samplesdir);

  my $samplespat = $samplesdir;
  $samplespat =~ s/([^-_:A-Za-z0-9])/\\$1/g;
  my $file;
  
  foreach $file (@TmpGlobal::site_files_found) {
    my $pretty = $file; $pretty =~ s,^${samplespat},\[samples\],g;

    if ($file =~ /layouts\.site/) {
      print CHOICE "    [x] (Site layouts for common sites)\n".
	      "\tFilename: $pretty\n\n";
      next;
    }

    my ($url, $name, $desc) = $self->ReadSiteForChoices($file);
    if (!defined $url) { next; }
    if (!defined $name) { $name = $url; }

    if (defined $desc) { $desc = "\t($desc)\n"; }
    else { $desc = ''; }

    my $chosen = ' ';
    if (defined $chosen_sites{$pretty}) { $chosen = 'x'; }

    print CHOICE "    [$chosen] $name\n\tURL: $url\n".
    		"\tFilename: $pretty\n$desc\n";
  }
  undef @TmpGlobal::site_files_found;

  close CHOICE or die "failed to write to site_choices file";

  my $edit;
  if (defined $ENV{'VISUAL'}) {
    $edit = $ENV{'VISUAL'};
  } elsif (defined $ENV{'EDITOR'}) {
    $edit = $ENV{'EDITOR'};
  } elsif (&Scoop::MyOS eq 'UNIX' && -x '/usr/bin/editor') {
    $edit = 'editor';		# Debian std, thanks to michael d. ivey
  } elsif (&Scoop::MyOS eq 'UNIX') {
    $edit = 'vi';
  } elsif (&Scoop::MyOS eq 'Win32') {
    $edit = 'notepad.exe';
  } elsif (&Scoop::MyOS eq 'Mac') {
    # REVISIT -- don't know what to do here ;)
    warn "\nIf you wish to choose which sites to scoop from a list\n".
    	"of pre-defined sites, stop this script now, edit the file\n".
	"$choicefile\n".
	"and re-run it.\n\n";
    return;
  }

  if (&Scoop::MyOS eq 'UNIX' && !-t STDOUT) {
    warn "\nNot running editor for site_choices.txt file, as sitescooper is\n".
      "not running interactively. Please edit this file at your\n".
      "convenience: $choicefile\n\n";

  } else {
    warn "Running editor for site_choices.txt file using command $edit...\n";
    system ($edit, $choicefile);

    if (($? >> 8) != 0) {
      die "The command failed. Please edit $choicefile\n".
	  "by hand and re-run sitescooper.\n\n";
    }
  }
}

# ---------------------------------------------------------------------------

sub ReadSiteForChoices {
  my $self = shift;
  my $file = shift;
  return if ($file =~ /(\.swp$|core|\.bak$|\~$|^#)/);
  return if (-d $file);

  open (IN, "<$file") || next;
  my ($url, $sitename, $desc);
  $url = $sitename = $desc = undef;
  while (<IN>) {
    s/
*$//g; s/#.*$//g;
    /^\s*Name:\s*(.*)$/ and ($sitename = $1), next;
    /^\s*Description:\s*(.*)$/ and ($desc = $1), next;
    /^\s*URL:\s*(.*)$/ and ($url = $1), next;
  }
  close IN;

  ($url, $sitename, $desc);
}

# ---------------------------------------------------------------------------

sub AddRegexpToSet {		# static
  my $regexp = shift;
  my $pat = shift;
  if (!defined ($regexp) || $regexp !~ /\)$/) {
    $regexp = "($pat)";
  } else {
    $regexp =~ s/\)$/|${pat})/g;
  }
  $regexp;
}

# ---------------------------------------------------------------------------

sub make_basic_dirs {
  my $self = shift;

  if (!-d $CF::tmpdir) {
    mkdir ($CF::tmpdir, 0777) || die "failed to mkdir '$CF::tmpdir'\n";
  }

  $CF::user_tmpdir = $CF::tmpdir.$CF::slash."sitescooper_$Scoop::userid";

  if (!-d $CF::user_tmpdir) {
    mkdir ($CF::user_tmpdir, 0777) || die "failed to mkdir '$CF::user_tmpdir'\n";
  }
}

sub make_dirs {
  my $self = shift;

  chdir ($CF::tmpdir) or die "cannot cd to $CF::tmpdir\n";

  if (-f "${CF::user_tmpdir}/cookies") {
    $Scoop::cookie_jar->load ("${CF::user_tmpdir}/cookies");
  }

  if ($CF::debug) {
    open (LOGFILE, "> ${CF::user_tmpdir}${CF::slash}log.txt");
    select LOGFILE; $| = 1; select STDOUT;
  }

  if ($CF::outdir eq '') { $CF::outdir = "${CF::user_tmpdir}${CF::slash}txt"; }
  if (!-d $CF::outdir) {
    mkdir ($CF::outdir, 0777) || die "failed to mkdir '$CF::outdir'\n";
  }

  $CF::cachedir = $CF::user_tmpdir.$CF::slash."cache"; $CF::newcachedir = $CF::cachedir;
  if (!-d $CF::cachedir) {
    mkdir ($CF::cachedir, 0777) || die "failed to mkdir '$CF::cachedir'\n";
  }

  if (defined $CF::sharedcache) {
    if (!-d $CF::sharedcache) {
      mkdir ($CF::sharedcache, 0777) || die "failed to mkdir '$CF::sharedcache'\n";
    }
  }

  $CF::alreadyseen = $CF::user_tmpdir.$CF::slash."already_seen.txt";
  $CF::newalreadyseen = $CF::alreadyseen;

  if ($CF::nowrite) {
    $CF::newcachedir = $CF::user_tmpdir.$CF::slash."new_cache";
    if (!-d $CF::newcachedir) {
      mkdir ($CF::newcachedir, 0777) || die "failed to mkdir '$CF::newcachedir'\n";
    }
    $CF::newalreadyseen = $CF::user_tmpdir.$CF::slash."new_already_seen.txt";
  }

  $CF::prcdir = $CF::user_tmpdir.$CF::slash."prc";

  # check for spaces on Win32 -- MakeDocW can't handle them!
  # Thx to wgoosey /at/ servtech.com for spotting this one.
  if ($CF::outputfilter eq 'makedoc') {
    if (&Scoop::MyOS eq 'Win32') {
      if ($CF::outdir =~ / /) {
	warn "

Warning: Sitescooper is installed in a directory containing spaces in the
filename. The MakeDocW conversion tool does not support this, so you may
need to move Sitescooper to another directory, e.g. C:\\Sitescooper, for
this conversion to work!  (This is a bug in MakeDOCW.exe.)

";
      }
    }
  }
}

# ---------------------------------------------------------------------------

sub expire_old_cache_files {
  my $self = shift;

  sub expire_cache { unlink if (-f $_ && -M $_ > $CF::expiry_days); }
  sub expire_shared_cache { unlink if (-f $_ && -M $_ > $CF::expiry_days); }

  my $stamp = $CF::cachedir.$CF::slash."last_clean.stamp";
  if (!defined (-M $stamp) || -M _ > $CF::expiry_days/2.0)
  {
    &Scoop::verbose ("Expiring cache files...");

    open (TOUCH, ">$stamp"); close TOUCH;
    find(\&expire_cache, $CF::cachedir);
    if (defined $CF::sharedcache) {
      find(\&expire_shared_cache, $CF::sharedcache);
    }
  }
}

# ---------------------------------------------------------------------------

sub read_state {
  my $self = shift;

  if ($CF::refresh == 0) {
    if (!open (IN, "< $CF::alreadyseen")) {
      &Scoop::verbose ("Cannot read $CF::alreadyseen, creating a new one");

    } else {
      my $url;
      my $mod;
      my $urlhost;
      while (<IN>) {
	/ lastmod=(\d+)$/;
	next if (!defined $1);
	$url = $`; $mod = $1;
	$Scoop::already_seen{$url} = 1;
	$Scoop::last_modtime{$url} = $mod+0;

	if ($url =~ m,http://(\S+?)/,) {
	  $urlhost = $1;
	  if (defined($Scoop::oldest_already_seen{$urlhost})
	      ? $Scoop::oldest_already_seen{$urlhost} > $mod : 1)
	  {
	    $Scoop::oldest_already_seen{$urlhost} = $mod;
	  }
	}
      }
      close IN;
    }
  }
}

# ---------------------------------------------------------------------------

sub generate_output_filenames {
  my $self = shift;
  my @sites = @_;
  my %already_done = ();
  my $url;
  my $scf;

  foreach $url (@sites) {
    next if ($url eq '');
    $scf = $AllSCFs::sites{$url};
    carp "url not defined in generate_output_filenames"
    			unless defined($url);
    carp "scf not defined in generate_output_filenames"
    			unless defined($scf);

    if (defined $scf->{req_cookie}) {
      ($TmpGlobal::req_cookie_host, $TmpGlobal::req_cookie_key)
      				= split (' ', $scf->{req_cookie});
      $TmpGlobal::gotit = 0;

      sub chk_for_reqd_cookie {
	if ($_[4] eq $TmpGlobal::req_cookie_host &&
	  	$_[1] eq $TmpGlobal::req_cookie_key)
	{
	  $TmpGlobal::gotit = 1;
	}
      }
      $Scoop::cookie_jar->scan (\&chk_for_reqd_cookie);

      if (!$TmpGlobal::gotit) {
	my $line = $scf->{site_defined_at};
	$line =~ s/^(.*):(.*?)$/"$1"/g;
	&Scoop::verbose ("Cookie from $TmpGlobal::req_cookie_host is not imported, not scooping $line.");
	$scf->{active} = 0;
      }

      undef $TmpGlobal::req_cookie_host;
      undef $TmpGlobal::req_cookie_key;
    }

    if (defined $scf->{req_env}) {
      my $envname = $scf->{req_env};
      if (!defined $ENV{$envname}) {
	my $line = $scf->{site_defined_at};
	$line =~ s/^(.*):(.*?)$/"$1"/g;
	&Scoop::verbose ("Env variable $envname is not set, not scooping $line.");
	$scf->{active} = 0;
      }
    }

    next unless ($scf->{active} == 1);

    my $sitename = $scf->{name};
    if (!defined $sitename) {
      $sitename = $url;
      $sitename =~ s/[\?\#].*$//;	# strip CGI params, or #anchors tho'
    }

    my $filedesc = $CF::filename_template;
    $filedesc =~ s/Site/${sitename}/g;
    $filedesc =~ s/Section//g;		# backwards compat

    # trim out dangerous chars (modified to allow most 8-bit stuff, thx to
    # <avatar /at/ deva.net>.
    $filedesc =~ s/[\s\;\*\$\%\!\&\<\>\|\?\'\"\`\]\[\{\}\:\\\/\000-\037]+/_/g;

    $filedesc =~ s/^[ _]+//g; $filedesc =~ s/[ _]+$//g;

    if (&Scoop::MyOS eq 'Mac') {
      # try to limit the filename to 32 characters
      $filedesc =~ s/^(.{26}).*$/$1/g;
    }

    my $outdir = $CF::outdir.$CF::slash.$filedesc;
    my $outidxfile;

    if ($self->{output_style} == $Scoop::OUT_HTML) {
      $outidxfile = $filedesc.'.html';
    } else {
      $outidxfile = $filedesc.'.txt';
    }

    next if (defined $already_done{$outdir});
    $already_done{$outdir} = 1;

    my $outtmp = $CF::outdir.$CF::slash.$filedesc.'.tmp';

    my $prctitle = $CF::prc_title;
    $prctitle =~ s/Site/${sitename}/g;
    $prctitle =~ s/Section//g;		# backwards compat
    $prctitle =~ s/^[ _]+//g; $prctitle =~ s/[ _]+$//g;

    my $scooper = new Sitescooper::Robot ($self,
    		$url, $AllSCFs::sites{$url},
		$outdir, $outidxfile);

    $scooper->{output_style} = $self->{output_style};
    $scooper->{outtmp} = $outtmp;
    $scooper->{site} = $url;
    $scooper->{sitename} = $sitename;
    $scooper->{prctitle} = $prctitle;

    $scooper->{cachedir} = $CF::cachedir;
    $scooper->{newcachedir} = $CF::newcachedir;
    $scooper->{sharedcachedir} = $CF::sharedcache;

    push (@Scoop::robots, $scooper);

    if ($CF::dumpprc) {
      $scooper->{syncfile} = $outtmp;		# reuse it!
    } else {
      $scooper->{syncfile} =
		  $CF::prcdir.$CF::slash.$filedesc.'.'.$CF::outputextn;
    }
  }
}

# ---------------------------------------------------------------------------

sub get_preloaded_page {
  my ($self, $url) = @_;
  my $resp = $self->{preloaded_responses}->{$url};

  if (defined $resp) {
    delete $self->{preloaded_responses}->{$url};
    $resp;
  } else {
    undef;
  }
}

# ---------------------------------------------------------------------------

sub get_all_sites {
  my $self = shift;
  my $convert_now = shift;
  if (!defined $convert_now) { $convert_now = 0; }

  # optimise by pre-getting front pages using parallel
  # requests, if possible.  Pre-get several pages in advance.
  my $ampreloading = ($self->{httpclient}->can_preload ()
  				&& !$CF::use_only_cache);

  if ($ampreloading) {
    &Scoop::dbg ("preloading front pages");
  }

  # now, run through the robots.  Preload the next few sites' front
  # pages while we process the current one.  Preload several in
  # advance, even though this will saturate the load queue, because
  # the less URLs we load from one site at a time the better.
  #
  my $PRELOAD_AHEAD = 3;
  my @robots = @Scoop::robots;

  my ($thisrobot);
  my @nextrobot = ();
  my $i;

  for ($i = 0; $i < $PRELOAD_AHEAD; $i++) {
    $nextrobot[$i] = shift @robots;
    if (defined $nextrobot[$i]) {
      $nextrobot[$i]->{convert_now} = $convert_now;
      if ($ampreloading) {
	$nextrobot[$i]->preload_front_page ($nextrobot[$i]->{url});
      }
    }
  }

  $i = $PRELOAD_AHEAD - 1;
  while (1) {
    $thisrobot = $nextrobot[0];
    shift @nextrobot;
    $nextrobot[$i] = shift @robots;

    if (defined $nextrobot[$i]) {
      $nextrobot[$i]->{convert_now} = $convert_now;
      if ($ampreloading) {
	$nextrobot[$i]->preload_front_page ($nextrobot[$i]->{url});
      }
    } else {
      $nextrobot[$i] = undef;
    }

    last unless (defined $thisrobot);	# out of sites
    $thisrobot->scoop_site();
  }
}

# ---------------------------------------------------------------------------

sub get_httpclient_queue_keys {
  my ($self) = @_;
  (keys %{$self->{httpclient_queue}});
}

sub add_httpclient {
  my ($self, $handler) = @_;
  $self->{httpclient_queue_new_idx}++;
  $self->{httpclient_queue}->{ $self->{httpclient_queue_new_idx} } = $handler;
}

sub remove_httpclient {
  my ($self, $handler) = @_;
  my $idx;
  foreach $idx (keys %{$self->{httpclient_queue}}) {
    my $ent = $self->{httpclient_queue}->{$idx};
    next unless (defined $ent && $ent eq $handler);
    delete $self->{httpclient_queue}->{$idx};
  }
}

sub httpclient_queue_to_string {
  my ($self) = @_;
  my $idx;
  my $ret = 'URL-handler queue: ';
  foreach $idx (keys %{$self->{httpclient_queue}}) {
    my $ent = $self->{httpclient_queue}->{$idx};
    $ret .= $ent->to_string().' ';
  }
  $ret =~ s/ $//;
  $ret;
}

# ---------------------------------------------------------------------------

sub convert_output {
  my $self = shift;
  my $robot = shift;
  my $scf = shift;

  carp "scf not defd" unless defined ($scf);
  carp "robot not defd" unless defined ($robot);

  return unless ($CF::use_convert_tool);

  my $syncfile = $robot->{syncfile};
  return unless defined $syncfile;

  my $outdir = $robot->{outdir};
  my $outidxfile = $robot->{outidxfile};
  my $prctitle = $robot->{prctitle};
  unlink $syncfile;

  my $idx = $outdir.$CF::slash.$outidxfile;

  if ($CF::outputfilter eq '__cat__') {
    open (IN, "< ".$idx);
    while (<IN>) { print STDOUT; }
    close IN;
    unlink $idx;
    return;
  }
  if ($CF::outputfilter eq '__path__') {
    print STDOUT "$idx\n";
    return;
  }

  my $cmd;
  if ($CF::outputfilter eq 'makedoc') {
    $cmd = "$CF::makedoc $CF::converter_args ".	
    		"\"$idx\" \"".$syncfile."\" '".$prctitle."'";

  } elsif ($CF::outputfilter eq 'plucker')
  {
    # make a quick HTTP exclusion list
    my $excludefile = $CF::user_tmpdir.$CF::slash."plucker_exclude.txt";
    open (OUT, ">". $excludefile) or die "cannot write to $excludefile\n";
    print OUT "0:-:http://.*\n";
    close OUT;

    # convert the full path into just the filename-style name of site
    my $sitename = $idx;
    $sitename =~ s,^.*[\/\\\:]([^\/\\\:]+)\.[^\.]+$,$1,;
 
    $cmd = "$CF::plucker $CF::pluckargs -f $sitename -N \"".$prctitle."\" ".
		     " ".$CF::converter_args." -H \"file:".$idx."\" ";
    if ($CF::fileperpage) {
      $cmd .= "-M 3 -E \"".$excludefile."\" ";
    } else {
      $cmd .= "-M 0 ";
    }
    $syncfile = $ENV{'HOME'}.$CF::slash.".plucker".$CF::slash.$sitename.".pdb";

  } elsif ($CF::outputfilter eq 'isilo') {
    if (&Scoop::MyOS eq 'Win32' && $CF::isilo =~ /isilow32/i) {
      $CF::isiloargs .= ' -u';		# doesn't support the other args AFAIK.
    }

    if ($CF::fileperpage) {
      $cmd = "$CF::isilo $CF::isiloargs ".
			"-i\"".$prctitle."\" ".
      			$CF::isilomultipageargs." ".
      			$CF::converter_args." ".
			"\"".$idx."\" ";
    } else {
      $cmd = "$CF::isilo $CF::isiloargs ".
			"-i\"".$prctitle."\" ".
      			$CF::converter_args." ".
			"\"".$idx."\" ";
    }

    # UNIX iSilo utils take the output filename as well; Win32
    # doesn't need it as it installs as it goes along.
    if (&Scoop::MyOS ne 'Win32') {
      $cmd .= " \"".$syncfile."\"";
    }

    # Win32 iSilo only takes the -u arg for the GUI version, not the
    # command line one. Strip the arg for the command-line converter.
    # Also add the output filename.
    if (&Scoop::MyOS eq 'Win32' && $cmd =~ /isiloc32/i) {
      $cmd =~ s/ -u / /g;
      $cmd .= " \"".$syncfile."\"";
    }

  } elsif ($CF::outputfilter eq 'richreader') {
    $cmd = "$CF::richreader $CF::richargs $CF::converter_args \"".$idx."\"";

  } elsif ($CF::outputfilter =~ /^cmd: /) {
    $cmd = $';
    $cmd =~ s/__SCOOPFILE__/${idx}/g;
    $cmd =~ s/__SYNCFILE__/${syncfile}/g;
    $cmd =~ s/__TITLE__/${prctitle}/g;
    $cmd =~ s/__ARGS__/${CF::converter_args}/g;

  } else {
    die "bad output filter $CF::outputfilter\n";
  }

  # substitute in the parameters taken from the site file
  # these should be set in the SCF::set_defaults method to avoid undefs.
  $cmd =~ s/__IMAGE_MAX_WIDTH__/ $scf->{image_max_width} /ge;

  my $keep_tmps = ($CF::debug || $CF::keep_tmps);

  if (&Scoop::MyOS ne 'Mac') {
    $self->add_cmd_dir_to_path ($cmd);
    &Scoop::verbose ("Running: $cmd");

    # cd to conversion dir for command
    my $realwd = getcwd; chdir $outdir;
    system $cmd;
    chdir $realwd;			# back again

    # output a newline, MakeDoc won't do it itself.
    if ($CF::outputfilter eq 'makedoc' && &Scoop::MyOS eq 'UNIX') {
      &Scoop::verbose ("\n");
    }

    if (($? >> 8) != 0)
    {
      # work around a bug in iSilo converter on Win32 -- it
      # reports failure even when the conversion went fine.
      # (TODO: check if still the case)
      if (&Scoop::MyOS ne 'Win32' ||
	    $CF::outputfilter ne 'isilo')
      {
	warn "command failed: $cmd\n";
	$Scoop::failed_to_cvt = 1;
	$keep_tmps = 1;
      }
    }

  } else
  {
    # system is broken on MacOS, so print the required command 
    #so it can be run easily from MPW shell
    if (!defined $Scoop::macos_system_warning_written) {
      warn "[Warning: not using the broken MacPerl system() call. ".
	    "You will need to\ncut and paste the command ".
	    "lines yourself!]\n\n";
      $Scoop::macos_system_warning_written = 1;
    }
    print $cmd, "\n";
    $keep_tmps = 1;
  }

  if (!$keep_tmps) {
    File::Path::rmtree ($outdir);		# don't keep .txt files around
  }

  if ($CF::dumpprc)
  {
    # If we're dumping, read in the generated file and write it to
    # STDOUT.
    open (IN, "< ".$syncfile);
    while (<IN>) { print STDOUT; }
    close IN;
    unlink $syncfile;

  } elsif (defined $CF::pilotinstdir)
  {
    # if installing to a dir, install it and output the filename.
    #
    my $filedesc = basename ($syncfile);
    my $instfile = $CF::pilotinstdir.$CF::slash.$filedesc;
    if ($syncfile ne $instfile) {
      move ($syncfile, $instfile) or
      	warn "failed to move $syncfile to $instfile\n";
    }
    print "Created: ".$instfile."\n";

  } elsif (defined $CF::pilotinstapp)
  {
    # If installing using an app, run it as appropriate, or for an
    # ***ADD_TO_MANIFEST*** install method, write the filename to
    # that manifest file.
    #
    if ($CF::pilotinstapp =~ /^\*\*\*ADD_TO_MANIFEST\*\*\* (.*)$/)
    {
      if (!open (OUT, ">> $1")) {
	warn "cannot write to $1\n";
      } else {
	print OUT $syncfile."\n"; close OUT;
      }

    } elsif ($CF::pilotinstapp eq '***USE_MODULE***') {
      my $instret = $CF::installer->install_file ($syncfile);
      &Scoop::dbg ("PDA::PilotInstall install_file returned ".
      		(defined $instret ? $instret : "undef"));
      if (defined $instret && $instret == 1) {
	unlink $syncfile;
      }

    } else {
      $cmd = "$CF::pilotinstapp $syncfile";
      $self->add_cmd_dir_to_path ($cmd);
      &Scoop::verbose ("Running: $cmd");
      system $cmd;

      if (($? >> 8) != 0) {
	warn "command failed: $cmd\n";
      } else {
	unlink $syncfile;
      }
    }
  } else
  {
    # just output the name of the file, in the sitescooper tmp directory,
    # for other apps that may want to collect these files and store
    # them somewhere.
    #
    print "Created: ".$syncfile."\n";
  }
}

sub add_cmd_dir_to_path {
  my $self = shift;
  local ($_);
  my $cmd = shift;

  # Perl on some Win32 platforms seems to require that the binary be
  # in the PATH.
  #
  if (&Scoop::MyOS eq 'Win32') {
    $_ = $cmd;
    if (!/[\\\/]/) { return; }	# foo arg ...
    if (/^\"([^\"]+)\"/) { $cmd = $1; }	# "C:\Program Files\foo.exe" arg ...
    elsif (/^(\S+)\s/) { $cmd = $1; }	# C:\windows\foo.exe arg ...
    else { $cmd = $_; }			# C:\windows\foo.exe

    $cmd =~ s,[\\/][^\\/]+\s*$,,g;		# trim the filename
    my $cmdpat = $cmd; $cmdpat =~ s,(\W),\\$1,g;	# escape funny chars

    if ($ENV{'PATH'} !~ /;${cmdpat}(;|$)/) {
      &Scoop::dbg ("Adding directory to command path: $cmd");
      my $path = $ENV{'PATH'} || $ENV{'Path'} || $ENV{'path'};
      $path .= ";$cmd"; $ENV{'PATH'} = $path;
    }
  }
}

# ---------------------------------------------------------------------------

sub write_state {
  my $self = shift;

  if (!$Scoop::failed_to_cvt) {
    # only write alreadyseen if the files converted successfully, otherwise
    # the user may lose some recent news due to a makedoc screwup.
    #
    my $towrite = '';
    my $now = time;
    my $twomonthsago = $now - (24*60*60*30*2);
    my $mod;
    my $urlhost;

    # keep the already-seen list small by cutting out old entries.  We
    # define "old entries" as (a) older than 2 months and (b) older than
    # the oldest link we saw in today's scooping run.
    #
    if (!$CF::refresh) {
      &Scoop::dbg ("trying to cut old entries from already-seen URL cache");

      foreach $_ (keys %Scoop::already_seen) {
	m,http://(\S+?)/,; $urlhost = $1; next unless defined ($urlhost);
	my $mod = $Scoop::last_modtime{$_};
	if (defined $mod && defined $Scoop::oldest_already_seen{$urlhost}) {
	  if ($twomonthsago > $mod &&
	    		$Scoop::oldest_already_seen{$urlhost} > $mod)
	  {
	    &Scoop::dbg ("stripping old entry: $_ lastmod=$mod (".&Scoop::time2datestr($mod).")");
	    next;
	  }
	}
	$towrite .= $_." lastmod=".(defined $mod ? $mod : $now)."\n";
      }

      if (open (OUT, "> $CF::newalreadyseen")) {
	print OUT $towrite;	# do it as one big atomic write, for safety
	close OUT || warn "Cannot rewrite $CF::newalreadyseen\n";
      } else {
	warn "Cannot rewrite $CF::newalreadyseen\n";
      }

    } else {
      # it's small enough -- so we can just append to it.
      &Scoop::dbg ("appending already-seen URLs to $CF::newalreadyseen");

      foreach $_ (@Scoop::seen_this_time) {
	$towrite .= $_." lastmod=".(defined $Scoop::last_modtime{$_}
		    ? $Scoop::last_modtime{$_} : $now)."\n";
      }

      if (open (OUT, ">> $CF::newalreadyseen")) {
	print OUT $towrite;	# do it as one big atomic write, for safety
	close OUT || warn "Cannot append to $CF::newalreadyseen\n";
      } else {
	warn "Cannot append to $CF::newalreadyseen\n";
      }
    }

    my ($from, $to);
    while (($from,$to) = each %CF::caches_to_rename) {
      &Scoop::dbg ("Saving new cache file: $to");
      rename ($from, $to) or warn ("rename $from -> $to failed\n");
    }
  }
}

# ---------------------------------------------------------------------------

sub disconnect {
  my $self = shift;

  if (&Scoop::MyOS eq 'Win32') {
    warn "disconnect on Win32 not implemented yet -- sorry.\n";

  } elsif (&Scoop::MyOS eq 'UNIX') {
    warn "disconnect: trying to kill pppd.\n";
    system ("killall pppd");	# probably won't work.

  } elsif (&Scoop::MyOS eq 'Mac') {
    eval q{
      use MacPerl;
      MacPerl::DoAppleScript("ppp disconnect");
    } or die "Cannot disconnect, MacPerl module not found: $@\n";
  }
}

# ---------------------------------------------------------------------------

sub parse_output_template {
  my $self = shift;

  $self->{html_main_page} = $self->{html_sub_page} = $self->{html_story} = undef;
  $self->{text_main_page} = $self->{text_sub_page} = $self->{text_story} = undef;

  # first read the defaults, then override them with the
  # -noheaders/-nofooters settings, then override those with user
  # template settings.
  #
  $self->parse_a_template ($self->{default_output_template});

  # strip out our tags if -noheaders is on. Be sure to leave in the
  # compulsory ones (__STORY_ANCHOR__, <html><head>, etc.)

  if ($CF::writeheader == 0)
  {
    $self->{html_main_page} =~ s/<h1>__SITE_TITLE__<\/h1>//is;
    $self->{text_main_page} =~ s/__SITE_TITLE__\n\n//is;

    $self->{html_story} =~ s/<hr>.*?<br>//is;
    $self->{html_story} =~ s/\[<a href=.*?<\/a>\]<br>//is;
    # leave in the <a name=...> tag though

    $self->{text_story} =~ s/__SITE_NAME__: __STORY_URL__\n\n//is;
  }

  if ($CF::writefooter == 0)
  {
    $self->{html_main_page} =~ s/<p><hr><font size=1>.*?<\/i><\/font>//is;
    $self->{text_main_page} =~ s/\(End of.*?__SITESCOOPER_HOME_URL__ \)\n//is;
  }

  # and now parse the user templates
  #
  if (defined $self->{output_template}) {
    $self->parse_a_template ($self->{output_template});
  }
}

sub parse_a_template {
  my $self = shift;
  local ($_) = shift;

  s/<htmlmainpage>(.*?)<\/htmlmainpage>//is and $self->{html_main_page} = $1;
  s/<htmlsubpage>(.*?)<\/htmlsubpage>//is and $self->{html_sub_page} = $1;
  s/<htmlstory>(.*?)<\/htmlstory>//is and $self->{html_story} = $1;

  s/<textmainpage>(.*?)<\/textmainpage>//is and $self->{text_main_page} = $1;
  s/<textsubpage>(.*?)<\/textsubpage>//is and $self->{text_sub_page} = $1;
  s/<textstory>(.*?)<\/textstory>//is and $self->{text_story} = $1;
}

# ---------------------------------------------------------------------------

sub warn_log {
  my $msg = join ('', @_); chomp $msg;
  &log ("Warning: ", $msg);
  &main::scoop_warn (@_);
}

sub die_log {
  my $msg = join ('', @_); chomp $msg;
  &log ("Fatal: ", $msg);
  &main::scoop_die (@_);
}

sub log {
  if (defined fileno LOGFILE) { print LOGFILE @_, "\n"; }
}

sub journal {
  my $self = shift;

  if (defined fileno Scoop::JOURNAL) {
    my $tag = shift;
    my $lines = join("", @_); $lines =~ s/^/$tag:\t/gm;
    print Scoop::JOURNAL $lines, "\n";
  }
}

sub dbg {
  if ($CF::debug != 0) {
    my $msg = "debug: ".join ('', @_); chomp $msg; &log ($msg);
    &main::dbg (@_);
  }
}

sub verbose {
  if ($CF::verbose) {
    my $msg = join ('', @_); chomp $msg; &log ($msg);
    &main::verbose (@_);
  }
}

# ---------------------------------------------------------------------------

sub sitewarn_file_line {
  my $fname = shift;
  $fname =~ s,^.*[\/\\:]([^\/\\:]+?):(\d+)?$,\"$1\" line $2,;
  &main::sitewarn ($fname, @_);
}

# ---------------------------------------------------------------------------

sub cleanexit {
  &main::cleanexit (@_);
}

sub read_test_support {
  my $file = shift;
  Scoop::dbg ("reading functional test support file $file");
  open (IN, "<$file") or die "no test file $file";
  while (<IN>) {
    s/\$\{(\S+?)\}/
	  defined($ENV{$1}) ? $ENV{$1} : "";
    /ge;
    if (/^Redirect: (\S+) (\S+)/) {
      $Scoop::func_test_redirect{$1} = $2;
    }
  }
  close IN;
}

sub AbsoluteURL {
  local ($_) = $_[1];
  my $fromurl = $_[0];

  s/^\s*\"//; s/\"\s*$//;	# trim quotes if necessary
  s/^\s*\'//; s/\'\s*$//;	# dodgy quotes
  s/^%22//; s/%22.*?$//;	# trim escaped quotes (!!)
  s/&amp;/&/g;			# HTML escapes are not supposed to be in URLs

  if (/^[^\/]+:/) {
    # experimental: if the URL begins with http:, assume it's already
    # absolute.
    return $_;
    #if (!/^(http|file):/) {
      ## non-HTTP urls get ignored; don't get URI::URL involved, it'll crash
      #return $_;
    #}
  }

  require URI::URL;
  my $url = new URI::URL ($_, $fromurl);
  $url->abs->as_string;
}

sub AddHostToURL {
  # a simpler form of AbsoluteURL, used for StoryURL lines.
  # this is necessary because the real thing will escape metacharacters
  # which screws up regexp patterns.

  local ($_) = $_[1];
  my $fromurl = $_[0];

  carp ("fromurl not defined in AddHostToURL") unless defined $fromurl;
  carp ("url not defined in AddHostToURL") unless defined $_;

  s/^"//; s/"$//;	# trim quotes if necessary
  $_ = &expand_url_magic ($_);	# allow [[MM]] etc. keywords in these patterns

  if (m,^[^/]+://,) {
    # do nothing, it's fully-qualified
  } elsif (m,^/,) {
    $fromurl =~ m,^([^/]+://[^/]+)/, and ($_ = $1.$_);
  }
  $_;
}

sub URLWithoutAnchor {
  my $url = shift;
  if (!defined $url) {
    warn ("undef in URLWithoutAnchor from: ".join(' ', caller()));
  }
  if ($url =~ /#/) {
    $`;
  } else {
    $url;
  }
}

sub URLAnchor {
  my $url = shift;
  if ($url =~ /#/) {
    "#".$';
  } else {
    "";
  }
}

sub mm_to_monthname {
  my @months = qw(x Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
  $months[$_[0]];
}

sub get_extra_date {
  my $time = shift;
  my ($x, $wday, $min, $hr);
  ($x,$min,$hr,$x,$x,$x,$wday,$x,$x) =
  		localtime(defined $time ? $time : time);
  my @days = qw(Sun Mon Tue Wed Thu Fri Sat);
  ($min, $hr, $days[$wday]);
}

sub get_date {
  my $time = $_[0];
  my ($x, $mday, $mon, $year);
  ($x,$x,$x,$mday,$mon,$year,$x,$x,$x) =
  		localtime(defined $time ? $time : time);
  $mon++; $year += 1900;
  ($mday, $mon, $year, &mm_to_monthname($mon));
}

sub time2datestr {
  my $time = shift;
  my ($dd, $mm, $year, $mon) = &get_date ($time);
  "$mon $dd $year";
}

sub match_url ($$$) {
  my ($cache, $url, $pat) = @_;

  if (!defined $url || !defined $pat) {
    warn ("undef in match_url from: ".join(' ', caller()));
    return undef;
  }

  my $re;
  $re = $cache->{$pat};
  if (!defined $re) { $re = $cache->{$pat} = qr{^${pat}(?:\#|$)}; }

  return $url =~ /$re/;
}

sub expand_url_magic {
  my $url = $_[0];
  if ($url !~ /\[\[/) { return $url; }

  local ($_);
  if (!defined $match_url_yyyy) {
    ($match_url_dd, $match_url_mm,
    $match_url_yyyy, $match_url_Mstr) = &get_date(undef);

    $match_url_yy = $match_url_yyyy;
    $match_url_yy =~ s/^\d\d//; # trim century
    $match_url_mm = "0$match_url_mm"
    			unless ($match_url_mm =~ /^..$/);
    $match_url_dd = "0$match_url_dd"
    			unless ($match_url_dd =~ /^..$/);
  }

  $url =~ s/\[\[YYYY\]\]/$match_url_yyyy/g;
  $url =~ s/\[\[YY\]\]/$match_url_yy/g;

  $url =~ s{\[\[MM([\+\-]\d+|)\]\]}{
    &offset_month($match_url_mm, $1);
  }ge;

  $url =~ s{\[\[M([\+\-]\d+|)\]\]}{
    &offset_month($match_url_mm, $1)+0;
  }ge;	# single-digit if poss

  $url =~ s{\[\[Mon([\+\-]\d+|)\]\]}{
    &mm_to_monthname (&offset_month($match_url_mm, $1));
  }ge;
  $url =~ s{\[\[mon([\+\-]\d+|)\]\]}{
    $_ = &mm_to_monthname (&offset_month($match_url_mm, $1));
    tr/A-Z/a-z/; $_;
  }ge;
  $url =~ s{\[\[MON([\+\-]\d+|)\]\]}{
    $_ = &mm_to_monthname (&offset_month($match_url_mm, $1));
    tr/a-z/A-Z/; $_;
  }ge;

  $url =~ s/\[\[DD\]\]/$match_url_dd/g;
  $url =~ s{\[\[D\]\]}{ $match_url_dd+0; }ge;	# single-digit if poss
  $url;
}

sub offset_month {
  my $mm = shift;
  my $offset = shift;
  if ($offset ne '') { $mm += $offset; }
  if ($mm < 1 || $mm > 12) { $mm = ((($mm-1)+12) % 12)+1; }
  $mm = "0$mm" unless ($mm =~ /^..$/);
  $mm;
}

# how a link should be marked up when we did not retrieve it, and
# it points to a page inside this site (but which did not match
# any of the retrieval URLs).
#
sub delink_unscooped_internal_link {
  my ($self, $url, $text) = @_;
  # "<u><!--internal-->".$text."</u>";
  "<u>".$text."</u>";
}

# how a link should be marked up when we did not retrieve it, and
# it points to a page outside this site.
#
sub delink_unscooped_external_link {
  my ($self, $url, $text) = @_;
  # "<u><!--external-->".$text."</u>";
  "<u>".$text."</u>";
}

# simple wrapper around Config and $^O to provide win/mac/unix
# differentiation, without worrying about which variant of UNIX
# it is.
#
sub MyOS {
  if (defined ($MY_OS)) { return $MY_OS; }

  # FIGURE OUT THE OS WE'RE RUNNING UNDER
  # Some systems support the $^O variable.  If not available then require()
  # the Config library.  [nicked from CGI.pm -- jmason]

  my $os;
  unless ($os) {
    unless ($os = $^O) {
      require Config;
      $os = $Config::Config{'osname'};
    }
  }

  if ($os=~/win/i) {
    $os = 'Win32';
  } elsif ($os=~/vms/i) {
    $os = 'VMS';
  } elsif ($os=~/mac/i) {
    $os = 'Mac';
  } elsif ($os=~/os2/i) {
    $os = 'OS2';
  } else {
    $os = 'UNIX';
  }
  $MY_OS = $os;
}

1;

#===========================================================================

# TODO:
#
# add a story inclusion/exclusion "grep" mode
# URLs at end like [1] this
# finish CGI support
#
#---------------------------------------------------------------------------
# vim:sw=2:tw=74:
