#===========================================================================
# 

package Sitescooper::Robot;

require Exporter;
use Carp;
use File::Basename;
use HTTP::Status;

use Sitescooper::StripTablesFilter;
use Sitescooper::LinksURLProcessor;
use Sitescooper::StoryURLProcessor;
use Sitescooper::ImageURLProcessor;

@ISA = qw(Exporter);
@EXPORT= qw();
$VERSION = "0.1";
sub Version { $VERSION; }

use strict;

sub new {
  my $class = shift; $class = ref($class) || $class;

  my $scoop = shift;
  my $url = shift;
  my $scf = shift;
  my $outdir = shift;
  my $outidxfile = shift;

  croak "scoop not defd" unless defined ($scoop);
  croak "url not defd" unless defined ($url);
  croak "scf not defd" unless defined ($scf);
  croak "outdir not defd" unless defined ($outdir);
  croak "outidxfile not defd" unless defined ($outidxfile);

  my $self = {
    'scoop'		=> $scoop,
    'url'		=> $url,
    'outdir'		=> $outdir,
    'outtmp'		=> $outdir.".tmp",	# default value
    'outidxfile'	=> $outidxfile,

    'output_style'	=> $Scoop::OUT_HTML,
    'syncfile'		=> undef,
    'prctitle'		=> undef,
    'sitename'		=> undef,
    'site'		=> undef,

    'cachedir'		=> undef,
    'newcachedir'	=> undef,
    'filekey'		=> undef,

    'convert_now'	=> 0,

    'sitekey'		=> undef,

    'warn_about_ext_links' => 0,	# turned on where necessary

    'already_tagged_for_downloading'  => { },
    'already_tried_download'          => { },

    'page_to_tmpfile'		=> { },
    'page_to_tmpfile_acc'	=> 0,

    'redirections'		=> { },
    'redirections_backwards'	=> { },
    'output_file'		=> { },

    'output_links_snarfed'  => { },
    'tmpfiles_assigned'	=> { },
    'anchors_assigned'  => { },
    'anchors_assigned_acc' => 0,

    'bookmark_urls'	=> [ ],
    'url_title'		=> { }
  };

  $self->{scf} = $scf;
  bless ($self, $class);
  $self->clear_page_tmpfiles();
  $self;
}

sub dont_convert {
  my $self = shift;
  if (!defined $self->{syncfile}) {
    croak "cannot block conversion without syncfile!\n";
  }
  delete $self->{syncfile};
}

sub scoop_site {
  local ($_);
  my $self = shift;
  my $url = $self->{url};

  if (!defined $self->{sitename}) { $self->{sitename} = $url; }

  # This apparently is needed on MacOS. Sounds unlikely, but there it
  # is...
  if (&Scoop::MyOS eq 'Mac') {
    my $parentdir = dirname($self->{outdir});
    if (!-d $parentdir) { mkdir ($parentdir, 0755); }
  }

  $_ = $self->{scf}->{site_defined_at}; /^(.*):/; my $site_file_name = $1;
  &Scoop::verbose ("SITE START: now scooping site \"$site_file_name\".");

  $self->dbg ("tmp dir: $self->{outtmp}, output dir: $self->{outdir}");

  (-d $self->{outtmp}) && File::Path::rmtree ($self->{outtmp});
  mkdir ($self->{outtmp}, 0755) || die "cannot mkdir $self->{outtmp}\n";

  $self->clear_page_tmpfiles();

  my $output_filename = $self->{outtmp}.$CF::slash.$self->{outidxfile};

  # evaluate perl code for this site.
  my $proc = $self->{scf}->get_story_param ('eval_code', $url);
  if (defined $proc) {
    my $skip_site = 0;
    if (!eval $proc."; 1;") {
      $self->sitewarn ("EvaluatePerl failed: $@");
    } else {
      if ($skip_site) {
	$self->dbg ("skip_site set, skipping this site.");
	next;
      }
    }
  }

  %{$self->{output_file}} = ();
  ${$self->{output_file}}{'MAIN'} = '';
  %{$self->{new_already_seen_age_cache_data}} = ();
  @{$self->{output_story_urls}} = ();
  $self->{stories_found} = 0;

  $self->{file_size} = 0;
  $self->{hit_file_size_limit} = 0;

  # work out while file size limit to use.
  {
    my $lim = undef;
    if (defined $CF::filesizelimit) { $lim = $CF::filesizelimit; }
    if (defined $self->{scf}->{size_limit} &&
	  $self->{scf}->{size_limit} < $lim)
    {
      $lim = $self->{scf}->{size_limit};
    }
    if (!defined $lim) { $lim = $CF::defaultfilesizelimit; }
    $self->{file_size_limit_in_use} = $lim;
  }

  my $upindex = $self->{current_story_index} = 0;

  &Scoop::set_got_intr_behaviour ('setflag');
  my $u;

  # if we're really running the internal unit tests, catch that here.
  if ($url eq 'file://__SITESCOOPER_INTERNAL_TESTS__/') {
    $self->dbg ("url is $url, running tests");
    eval '
      require Sitescooper::UnitTests;
      Sitescooper::UnitTests::run_top_level ($self);
    1; ' or die $@;
  }

  my @urls = $self->grep_unseen_urls ($url,
		 split (' ', $self->{scf}->{top_level_extra_urls}));
  foreach $u (@urls) {
    # if we were interrupted, clear the flag and go on
    if ($Scoop::got_intr_flag) { &Scoop::set_got_intr_behaviour ('setflag'); }
    $self->add_page_tmpfile ($self->{outtmp}.$CF::slash.$self->{outidxfile}, $u);

    if ($self->{scf}->{levels} >= 0) {
      $self->download_front_page (undef, $u, $self->{scf}->{levels}, $upindex);
    } else {
      $self->download_story_page (undef, $u, 1, $upindex);
    }
  }

  # Now go through any additional URLs at the different levels, starting
  # at the highest level and working down.
  my $lev;
  for ($lev = $self->{scf}->{levels}; $lev >= 0; $lev--) {
    next unless (defined $self->{scf}->{links_extra_urls}->{$lev});
    my @urls = $self->grep_unseen_urls
		 (split (' ', $self->{scf}->{links_extra_urls}->{$lev}));
    foreach $u (@urls) {
      # if we were interrupted, clear the flag and go on
      if ($Scoop::got_intr_flag) { &Scoop::set_got_intr_behaviour ('setflag'); }
      $self->add_page_tmpfile
      		($self->{outtmp}.$CF::slash.$self->{outidxfile}, $u);
      $self->download_front_page (undef, $u, $lev, $upindex);
    }
  }

  # and finally the extra story URLs
  foreach $u (split (' ', $self->{scf}->{story_extra_urls})) {
    # if we were interrupted, clear the flag and go on
    if ($Scoop::got_intr_flag) { &Scoop::set_got_intr_behaviour ('setflag'); }
    $self->add_page_tmpfile
    		($self->{outtmp}.$CF::slash.$self->{outidxfile}, $u);
    $self->download_story_page (undef, $u, 1, $upindex);
  }

  # allow the retrieval to finish up.
  1 while $self->handler_run_queue();
  &Scoop::set_got_intr_behaviour ('exit');

  # kld: addition for image-only sites
  # jm -- TODO:  needs checking, this. Seems OK though
  if (defined $self->{scf}->{image_only_site} &&
	$self->{scf}->{image_only_site} == 1 &&
	$self->{stories_found} == 0)
  {
    $self->dbg ("$self->{prctitle}: forcing conversion for image-only site");
    $self->{stories_found} = 1;
  }

  if ($self->{stories_found} > 0 &&
    		$self->{stories_found} >= $self->{scf}->{min_pages})
  {
    &Scoop::verbose ("$self->{prctitle}: $self->{stories_found} ".
	    "stories downloaded (".
	    sprintf ("%3.1f", $self->{file_size}/1024).
	    " K uncompressed).");

    my $ind = $self->{current_story_index};
    my $ofkey;
    foreach $ofkey (keys %{$self->{output_file}})
    {
      # convert sitescooper navigation links: [<<][^][>>]
      my $story = ${$self->{output_file}}{$ofkey};

      # trim off the first and last ones anyway
      $story =~ s/\[<a href=\"__SITESCOOPER_STORY_(-1|${ind})\">.*?<\/a>\]//g;

      # and run through the rest
      my $i; for ($i = 0; $i < $ind; $i++) {
	next unless (defined ${$self->{output_story_urls}}[$i]);
	$story =~
	s/\"__SITESCOOPER_STORY_${i}\"/\"${$self->{output_story_urls}}[$i]\"/g;
      }

      # rewrite links that turned out to redirect to another URL
      $story = $self->rewrite_redirected_links ($story);

      # remove stray links
      ${$self->{output_file}}{$ofkey} = $self->remove_external_links ($story);
    }

    $self->write_main_output_file ($output_filename);

    if (!$CF::fileperpage) {
      if ($CF::dump) {
	open (IN, "<".$self->{outtmp}.$CF::slash.$self->{outidxfile});
	while (<IN>) { print STDOUT; }
	close IN; File::Path::rmtree ($self->{outtmp});

	# ensure we do not try to convert it later
	$self->dont_convert();

      } else {
	File::Path::rmtree ($self->{outdir});
	rename ($self->{outtmp}, $self->{outdir});
	if ($self->{convert_now}) {
	  $self->{scoop}->convert_output($self, $self->{scf}, $url);
	}
      }

    } else {
      foreach $_ (keys %{$self->{output_file}}) {
	next if ($_ eq 'MAIN');
	$self->write_sub_output_file ($_);
      }

      File::Path::rmtree ($self->{outdir});
      rename ($self->{outtmp}, $self->{outdir});

      if ($CF::dump) {
	# print the path to the index page
	print $self->{outdir}.$CF::slash.$self->{outidxfile}."\n";
	# ensure we do not try to convert it later
	$self->dont_convert();

      } else {
	if ($self->{convert_now}) {
	  $self->{scoop}->convert_output($self, $self->{scf}, $url);
	}
      }
    }

    $self->dbg ("output dir: ".$self->{outdir});
    $self->dbg ("output index: ".
    		$self->{outdir}.$CF::slash.$self->{outidxfile});

    my ($from, $to);
    while (($from,$to) = each %{$self->{new_already_seen_age_cache_data}}) {
      $self->dbg ("Saving new already_seen age cache entry: $from => $to ".
	      "(".&Scoop::time2datestr($to).")");
      $Scoop::oldest_already_seen{$from} = $to;
    }

  } else {
    close OUTFILE;
    if ($self->{stories_found} > 0 && $self->{scf}->{min_pages} < $self->{stories_found})
    {
      &Scoop::verbose ("$self->{prctitle}: not enough stories, ignoring.");
    } else {
      &Scoop::verbose ("$self->{prctitle}: no new stories, ignoring.");
    }
    $self->dbg ("(Not setting already_seen age cache since no links were followed)");
    $self->dont_convert();
    File::Path::rmtree ($self->{outtmp});
  }

  &Scoop::verbose ("SITE END: done scooping site \"$site_file_name\".");
  delete ${$self->{output_file}}{'MAIN'};
}

# ---------------------------------------------------------------------------

sub write_main_output_file {
  my ($self, $output_filename) = @_;
  open (OUTFILE, "> $output_filename")
		  or die "Failed to create $output_filename\n";

  my $tmpl = $self->{scoop}->{html_main_page};
  if ($self->writing_doc || $self->writing_text) {
    $tmpl = $self->{scoop}->{text_main_page};
  }

  print OUTFILE $self->rewrite_template ($tmpl, {
	'__MAIN_BODY__' => ${$self->{output_file}}{'MAIN'}
      });

  if ($self->writing_doc) {
    print OUTFILE "<$Scoop::bookmark_char>\n";
  }

  close OUTFILE or warn "Failed to write to $output_filename";
}

sub write_sub_output_file {
  my ($self, $filename) = @_;
  open (OUTFILE, "> $filename")
		  or die "Failed to create $filename\n";

  my $tmpl = $self->{scoop}->{html_sub_page};
  if ($self->writing_doc || $self->writing_text) {
    $tmpl = $self->{scoop}->{text_sub_page};
  }

  print OUTFILE $self->rewrite_template ($tmpl, {
	'__SUB_BODY__' => ${$self->{output_file}}{$filename}
      });

  if ($self->writing_doc) {
    print OUTFILE "<$Scoop::bookmark_char>\n";
  }

  close OUTFILE or warn "Failed to write to $filename";
}

# ---------------------------------------------------------------------------

sub rewrite_template {
  my $self = shift;
  my $str = shift;
  my $vars = shift;

  my ($k, $v);
  while (($k, $v) = each %$vars) {
    next unless defined $k;
    $v ||= '';
    $str =~ s/${k}/${v}/gs;
  }

  $str =~ s/__SITE_TITLE__/$self->{prctitle}/gs;
  $str =~ s/__SITE_NAME__/$self->{sitename}/gs;
  $str =~ s/__SITESCOOPER_HOME_URL__/${Scoop::home_url}/gs;
  $str =~ s/__BOOKMARK_CHAR__/${Scoop::bookmark_char}/gs;
  $str =~ s/__ISILO_BOOKMARK_LINKS__/ $self->gen_isilo_bookmark_links(); /ges;

  my $rights = '';
  if (defined $self->{scf}->{rights}) { $rights = $self->{scf}->{rights}; }
  $str =~ s/__SITE_RIGHTS__/${rights}/gs;

  $str;
}

sub gen_isilo_bookmark_links {
  my $self = shift;
  my $url;

  if (!$CF::include_isilo_bookmarks) {
    return '';
  } # else

  my $str = "<X-ISILO TYPE=\"DOCUMENT\">";
  foreach $url (@{$self->{bookmark_urls}}) {
    my $title = $self->{url_title}{$url};

    # This should result in something like:
    # <A HREF="toc.htm#Table Of Contents">
    # <A HREF="page1.htm#First Page">
    # <A HREF="#Second Page">
    # (see http://www.isilo.com/support/html.htm#bookmarks )
    #
    if ($CF::fileperpage) {
      $str .= "<A HREF=\"".
      		$self->href_to_multipage_href($url). "#".$title."\">\n";
    } else {
      $str .= "<A HREF=\"#".$title."\">\n";
    }
  }

  $str . "</X-ISILO>\n";
}

# ---------------------------------------------------------------------------

sub download_front_page {
  my ($self, $refer, $url, $level, $upindex) = @_;

  my $handler = new Sitescooper::LinksURLProcessor ($self->{scoop},
  	$self, $self->{scf}, $refer, $url, $level, $upindex);
  $self->handler_start ($handler);
}

# ---------------------------------------------------------------------------

sub download_story_page {
  my ($self, $refer, $url, $is_dynamic_html, $upindex) = @_;

  my $handler = new Sitescooper::StoryURLProcessor ($self->{scoop},
  	$self, $self->{scf}, $refer, $url, $is_dynamic_html, $upindex);
  $self->handler_start ($handler);
}

# ---------------------------------------------------------------------------

sub download_image {
  my ($self, $refer, $url, $tag) = @_;

  my $handler = new Sitescooper::ImageURLProcessor ($self->{scoop},
  	$self, $self->{scf}, $refer, $url, $tag);
  $self->handler_start ($handler);
  $handler->{relative};
}

# ---------------------------------------------------------------------------

sub re_request_page {
  my ($self, $handler) = @_;
  $self->handler_start ($handler);
}

# ---------------------------------------------------------------------------

sub handler_start {
  my ($self, $handler) = @_;

  my @keys;
  for (@keys = $self->{scoop}->get_httpclient_queue_keys();
  	$#keys + 1 >= $self->{scoop}->{httpclient}->get_max_active_requests();
  	@keys = $self->{scoop}->get_httpclient_queue_keys())
  {
    $self->dbg ("queue needs to empty before url-handler can start: ".
    	$handler->to_string());
    $self->dbg ($self->{scoop}->httpclient_queue_to_string());

    # spin our wheels until a queue entry leaves the queue
    my $runqstatus;
    1 while (defined($runqstatus = $self->handler_run_queue_once())
    			&& $runqstatus == 0);
  }

  if ($handler->start_get()) {
    $self->dbg ("adding url-handler to queue: ".$handler->to_string());
    $self->{scoop}->add_httpclient ($handler);
  }
  $self->dbg ($self->{scoop}->httpclient_queue_to_string());
}

# ---------------------------------------------------------------------------

sub preload_front_page {
  my ($self, $url) = @_;

  my $handler = new Sitescooper::PreloadURLProcessor ($self->{scoop},
  	$self, $self->{scf}, undef, $url);
  $self->handler_start ($handler);
}

sub finish_preload {
  my ($self) = @_;

  $self->handler_run_queue_once();
}

# ---------------------------------------------------------------------------
# run the handler queue. Keep trying until all handlers have finished
# retrieving pages.
#
sub handler_run_queue {
  my ($self) = @_;
  local ($_);
  my $firsttime = 1;

  $self->dbg ("waiting for queue to empty: ".
  			$self->{scoop}->httpclient_queue_to_string());

  while (1) {
    if ($firsttime == 0) {
      # we're going in circles; sleep a little so we don't eat
      # 100% cpu while we're doing this.
      select (undef, undef, undef, 0.25);
    } else {
      $firsttime = 0;
    }

    if (!defined ($self->handler_run_queue_once())) {
      $self->dbg ("queue now completely empty");
      last;
    }

    my @keys = $self->{scoop}->get_httpclient_queue_keys();
    my $stillone = 0;
    my $idx;
    foreach $idx (sort @keys) {
      $_ = $self->{scoop}->{httpclient_queue}->{$idx};
      next unless defined $_;
      if ($_->{robot} == $self) {
	$self->dbg ("still one of our httpclients in queue: ".
  			$self->{scoop}->httpclient_queue_to_string());
	$stillone = 1; last;
      }
    }
    next if $stillone;

    $self->dbg ("queue now empty of our httpclients: ".
  			$self->{scoop}->httpclient_queue_to_string());
    last;
  }

  undef;
}

# ---------------------------------------------------------------------------

sub url_is_in_queue {
  my ($self, $url) = @_;
  local ($_);
  my @keys = $self->{scoop}->get_httpclient_queue_keys();
  my $stillone = 0;
  my $idx;

  foreach $idx (@keys) {
    $_ = $self->{scoop}->{httpclient_queue}->{$idx};
    next unless defined $_;
    if ($_->{robot} == $self && $_->{url} eq $url) {
      $self->dbg ("URL already loading: ".
		      $self->{scoop}->httpclient_queue_to_string());
      return 1;
    }
  }
  return 0;
}

# ---------------------------------------------------------------------------

sub handler_run_queue_once {
  my ($self) = @_;
  local ($_);
  my @hdlrs = ();
  my ($idx);

  # first of all, get the keys and sort them, so we (to a reasonable
  # degree) get the pages in order.
  my @keys = $self->{scoop}->get_httpclient_queue_keys();
  foreach $idx (sort @keys) {
    $_ = $self->{scoop}->{httpclient_queue}->{$idx};
    next unless defined $_;
    push (@hdlrs, $_);
  }
  if ($#hdlrs < 0) { return undef; }		# nothing left to get

  @hdlrs = Sitescooper::URLProcessor::get_ready_handlers (@hdlrs);
  if ($#hdlrs == 0 && !defined $hdlrs[0]) { return undef; }	# none ready - yet

  my $handlers_done = 0;
  my $h;
  foreach $h (@hdlrs) {
    if (!$h->run) {
      $self->dbg ("url-handler done: ".$h->to_string());
      $self->{scoop}->remove_httpclient ($h);
      $handlers_done++;
    }
  }
  $handlers_done;
}

# ---------------------------------------------------------------------------

sub remove_an_ext_link {
  my $self = shift;

  my ($link, $text, $ahref, $posthref) = @_;

  if (!$self->writing_html) {
    return $text;
  }

  if ($CF::nolinkrewrite) { goto gotit; }
  if (defined (${$self->{output_links_snarfed}}{$link})) { goto gotit; }
  if (!$CF::fileperpage && $link =~ /__HASH__/
	&& defined (${$self->{output_links_snarfed}}{$`})) { goto gotit; }
  if ($CF::fileperpage && $link =~ /#/
    	&& defined (${$self->{output_links_snarfed}}{$`})) { goto gotit; }

  $self->dbg ("Removing non-snarfed link: $link (\"$text\")");
  return $self->{scoop}->delink_unscooped_external_link ($link, $text);

gotit:
  $ahref.$link.$posthref.$text."</a>";
}

# ---------------------------------------------------------------------------

sub remove_external_links {
  my $self = shift;

  local ($_) = $_[0];

  $self->dbg2 ("all links found: ".join(' ',
  			sort keys %{$self->{output_links_snarfed}}));

  # jm: imposed limit of 200 chars on link text to stop runaway links
  s/(<a\s+[^>]*href\s*=\s*\")([^\"]+)(\"[^>]*?>)(.{0,200}?)<\/a>/
	  $self->remove_an_ext_link ($2, $4, $1, $3);
      /gies;
  s/(<a\s+[^>]*href\s*=\s*\')([^\']+)(\'[^>]*?>)(.{0,200}?)<\/a>/
	  $self->remove_an_ext_link ($2, $4, $1, $3);
      /gies;

  # fix REAL external links so they're now active and valid
  s/HREF_EXTERNAL/href/gs;

  $_;
}

# ---------------------------------------------------------------------------

sub rewrite_redirected_links {
  my $self = shift;
  local ($_) = $_[0];
  my $oldurl;

  foreach $oldurl (keys %{$self->{redirections}}) {
    my $newurl = $self->{redirections}{$oldurl};
    &Scoop::dbg ("fixing redirection: $oldurl => $newurl");

    if ($CF::fileperpage) {
      $oldurl = $self->href_to_multipage_href ($oldurl);
      $newurl = $self->href_to_multipage_href ($newurl);
    } else {
      $oldurl = "#".$self->href_to_singlepage_href ($oldurl);
      $newurl = "#".$self->href_to_singlepage_href ($newurl);
    }

    s/href=\s*[\"\']${oldurl}[\"\']/href=\"${newurl}\"/g;
  }
  $_;
}

# ---------------------------------------------------------------------------

sub check_for_oldest {
  my $self = shift;
  my $url = shift;

  my $fullurl = $url; $url = &Scoop::URLWithoutAnchor ($url);
  $url =~ m,http://(\S+?)/,i; my $urlhost = $1;
  return unless defined ($urlhost);

  #$self->dbg ("checking to see if $url is oldest at its site: modtime=".
  	#(defined $Scoop::last_modtime{$url} ? $Scoop::last_modtime{$url} : "unknown)"));

  if (defined $Scoop::last_modtime{$url}) {
    if (defined(${$self->{new_already_seen_age_cache_data}}{$urlhost})
	? ${$self->{new_already_seen_age_cache_data}}{$urlhost} > $Scoop::last_modtime{$url} : 1)
    {
      $self->dbg ("oldest link seen at $urlhost $url: modtime=".
		$Scoop::last_modtime{$url}." (".
		&Scoop::time2datestr($Scoop::last_modtime{$url}).")");
      ${$self->{new_already_seen_age_cache_data}}{$urlhost} = $Scoop::last_modtime{$url};
    }
  }
}

# ---------------------------------------------------------------------------

sub generic_cache_fname {
  my $dir = shift;
  my $url = shift;

  if (!defined $dir) { return undef; }

  $url = &Scoop::URLWithoutAnchor ($url);		# trim #anchors
  
  my $site;
  my $path;
  if ($url =~ m,^([\w]+://[^/]+)/(.+)$,) {
    $site = $1; $path = $2;
  } elsif ($url =~ m,^([\w]+):///(.+)$,) {	# eg. file:///some/file
    $site = $1; $path = $2;
  } else {
    $site = $url; $path = '_'; $site =~ s/\/$//;
  }
  $site =~ s,^http://,,i; $site =~ s,^www\.,,i;	# trim common stuff
  $site =~ s/[^-_A-Za-z0-9]/_/g;
  $path =~ s/[^-_A-Za-z0-9]/_/g;

  if ($CF::use_hashes_for_cache_filenames) {
    # try to limit the filename by trimming the start and adding the
    # hash value at the beginning instead. Let's hope this is not
    # too prone to collisions...
    if ($site =~ /^(.+)(.{16})$/) {
      $site = sprintf ("%4x_%s", unpack ("%16C*", $1), $2);
    }
    if ($path =~ /^(.+)(.{16})$/) {
      $path = sprintf ("%4x_%s", unpack ("%16C*", $1), $2);
    }
  }

  $site = $dir.$CF::slash.$site;
  if (!-d $site) {
    mkdir ($site, 0777) or die "failed to mkdir '$site'\n";
  }
  $site .= $CF::slash.$path;
  $site;
}

sub cachefilename {
  my $self = shift; &generic_cache_fname ($self->{cachedir}, $_[0]);
}
sub sharedcachefilename {
  my $self = shift; &generic_cache_fname ($self->{sharedcachedir}, $_[0]);
}
sub newcachefilename {
  my $self = shift; &generic_cache_fname ($self->{newcachedir}, $_[0]);
}

sub get_cached_page {
  my $self = shift;
  my $url = shift;
  my $is_diff_page = shift;
  my $cachefile = $self->cachefilename ($url);

  if (!defined $cachefile) { return undef; }

  # if -refresh is on, do not return any cached pages.
  # if it and -fromcache are both on, *do* return them.
  if ($CF::refresh && !$CF::use_only_cache) { return undef; }

  if (open (IN, "< $cachefile")) {
    binmode IN; my $cachedpage = join ('', <IN>); close IN;
    my $lastmod;
    if ($cachedpage =~ s/^<!-- lastmod: (\d+) -->\n//) { $lastmod = $1+0; }
    ($cachedpage, $lastmod);
  } else {
    undef;
  }
}

sub get_cached_page_for_diff {
  my $self = shift; $self->get_cached_page (@_, 1);
}

# ---------------------------------------------------------------------------
 
sub cache_page {
  my ($self, $url, $redir_from, $page, $cachelater) = @_;

  $url = &Scoop::URLWithoutAnchor ($url);
  my $cachefile = $self->newcachefilename ($url);

  # if this page is the latest version of a diffed page, don't cache it
  # immediately, as it will mean lost stories if we're interrupted.
  # Instead save the filename for renaming when the run finishes.
  if (defined $cachelater && $cachelater == 1) {
    my $tmpname = $cachefile.'.tmp';
    $CF::caches_to_rename{$tmpname} = $cachefile;
    $cachefile = $tmpname;
  }

  open (C1OUT, "> $cachefile"); binmode C1OUT; print C1OUT $page; close C1OUT;

  if (defined $CF::sharedcache) {
    $cachefile = $self->sharedcachefilename ($url);
    open (C2OUT, "> $cachefile"); binmode C2OUT;
    if (defined $Scoop::last_modtime {$url}) {
      # cache the last-modified time of this page as well.
      print C2OUT "<!-- lastmod: ",$Scoop::last_modtime{$url}," -->\n";
    }
    print C2OUT $page; close C2OUT;
  }

  if (defined $redir_from) {
    # TODO: make caching redirected pages more efficient
    $self->cache_page ($redir_from, undef, $page, $cachelater);
  }
}

 
sub cache_page_later { &cache_page ($_[0], $_[1], $_[2], $_[3], 1); }

# ---------------------------------------------------------------------------

sub write_as_story {
  local ($_);
  my $self = shift;
  my ($is_front, $url, $page, $headline, $upindex) = @_;

  my $fullurl = $url;
  $url = &Scoop::URLWithoutAnchor ($url);

  # provide a shorter version of the URL for palm-sized screens. Split
  # at both / and & signs to ensure CGI urls work too.
  my $shorturl = $url;
  if (length $shorturl > 35) { $shorturl =~ s/^.*[\/\&](.{20,35})$/...\/$1/; }

  my $one_page_anchor = '';
  my $one_page_anchor_tag = '';
  if ($self->writing_html && !$CF::fileperpage) {
    $one_page_anchor = $self->href_to_singlepage_href ($url);
    $one_page_anchor_tag = "<a name=\"$one_page_anchor\">";
  }

  if (defined $headline) {
    $self->set_url_title ($url, $headline);
  }

  # for text output, split at newlines appropriately.
  if ($self->writing_text || $self->writing_doc) {
    my @page = (split (/\n/, $page));
    $page = '';
    foreach $_ (@page) {
      if ($self->writing_text) {
	# wrap each line after 70 columns
	while (s/^(.{70}\S*)\s+//) {
	  $page .= $1."\n";
	}
      }
      $page .= $_."\n";
    }
  }

  my $tmpl = $self->{scoop}->{html_story};
  if ($self->writing_doc || $self->writing_text) {
    $tmpl = $self->{scoop}->{text_story};
  }

  my $hdlineank = '';
  if ($CF::include_isilo_bookmarks && defined $headline) {
    $_ = $headline; s/\"/\'/g;
    $hdlineank = "<a name=\"$_\">";
  }

  my $outtext = $self->rewrite_template ($tmpl, {
	'__STORY_TEXT__' => $page,
	'__STORY_ANCHOR__' => $one_page_anchor_tag,
	'__STORY_URL__' => $url,
	'__SHORT_URL__' => $shorturl,
	'__STORY_PREV_LINK__' =>
		'__SITESCOOPER_STORY_'.($self->{current_story_index}-1),
	'__STORY_UP_LINK__' =>
		'__SITESCOOPER_STORY_'.$upindex,
	'__STORY_NEXT_LINK__' =>
		'__SITESCOOPER_STORY_'.($self->{current_story_index}+1),
	'__HEADLINE__' => $headline,
	'__ISILO_HEADLINE_ANCHOR__' => $hdlineank
      });

  if ($CF::fileperpage) {
    my $fname = $self->href_to_multipage_filename ($url);
    my $relative = $self->href_to_multipage_href ($url);

    if ($is_front && $fullurl eq $self->{url}) {
      # this is the front page, just append it to the index file
      ${$self->{output_file}}{'MAIN'} .= $outtext;
      ${$self->{output_story_urls}}[$self->{current_story_index}] = $relative;

    } elsif ($self->writing_html)
    {
      my $tmpl = $self->{scoop}->{html_sub_page};
      if ($self->writing_doc || $self->writing_text) {
	$tmpl = $self->{scoop}->{text_sub_page};
      }

      ${$self->{output_file}}{$fname} = $self->rewrite_template ($tmpl, {
	    '__SUB_BODY__' => $outtext,
	    '__HEADLINE__' => $headline,
	  });
      ${$self->{output_story_urls}}[$self->{current_story_index}] = $relative;

      if ($fullurl ne $url) {
	my $ffname = $self->href_to_multipage_filename ($fullurl);
	my $frelative = $self->href_to_multipage_href ($fullurl);
	$self->add_snarfed_link ($ffname);
	$self->add_snarfed_link ($frelative);
      }

    } else {
      ${$self->{output_file}}{$fname} .= $outtext;
    }

  } else {
    # this is single-page output mode.
    ${$self->{output_file}}{'MAIN'} .= $outtext;

    if ($self->writing_html) {
      $self->add_snarfed_link ('#'.$one_page_anchor);
      ${$self->{output_story_urls}}[$self->{current_story_index}]
      						= '#'.$one_page_anchor;

      if ($one_page_anchor =~ /__HASH__/) {
	$self->add_snarfed_link ('#'.$`);
      }
    }
  }

  $self->{current_story_index}++;
  $self->add_snarfed_link ($url);
  if ($fullurl ne $url) { $self->add_snarfed_link ($fullurl); }
  $self->up_file_size ($url, length($outtext), "story");
  $self->{stories_found}++;

  if ($CF::storylimit > 0 && $self->{stories_found} >= $CF::storylimit) {
    &Scoop::verbose ("over story limit, stopping this site.");
    $self->{hit_file_size_limit} = 1;
  }
}

# ---------------------------------------------------------------------------

sub up_file_size {
  my $self = shift;
  my $url = shift;
  my $siz = shift;
  my $typetxt = shift;

  $self->{file_size} += $siz;
  $self->dbg ("$typetxt written, ".
  	($self->{file_size}/1024)." K, limit ".
	$self->{file_size_limit_in_use}." K");

  if ($self->{file_size}/1024 >= $self->{file_size_limit_in_use}) {
    $self->{hit_file_size_limit} = 1;
  }
}

# ---------------------------------------------------------------------------

sub add_snarfed_link {
  my ($self, $mainurl) = @_;
  my $url;

  my $redir_from = $self->{redirections_backwards}{$mainurl};
  $redir_from ||= '';

  foreach $url ($mainurl, split (' ', $redir_from)) {
    if ($CF::debug > 1) {
      my ($class, $file, $line) = caller();
      $self->dbg2 ("Tracking snarfed link: $url (at $class:$line)");
    }

    ${$self->{output_links_snarfed}}{$url} = 1;

    if ($CF::fileperpage) {
      my $relative = $self->href_to_multipage_href ($url);
      ${$self->{output_links_snarfed}}{$relative} = 1;
    } else {
      my $anchor = $self->href_to_singlepage_href ($url);
      ${$self->{output_links_snarfed}}{"#".$anchor} = 1;
    }
  }
}

# ---------------------------------------------------------------------------

sub clear_page_tmpfiles {
  my $self = shift;
  %{$self->{page_to_tmpfile}} = ();
  $self->{page_to_tmpfile_acc} = 0;
}

sub add_page_tmpfile {
  my $self = shift;
  my $fname = shift;
  my $url = shift;
  $self->{page_to_tmpfile}->{$url} = $fname;
  $self->{tmpfiles_assigned}->{$fname} = 1;

  if ($CF::debug > 1) {
    my ($class, $file, $line) = caller();
    $self->dbg2 ("assigned page file: $url = $fname: (at $class:$line)");
  }
}

# ---------------------------------------------------------------------------

sub href_to_multipage_filename {
  my $self = shift;
  my $url = shift;
  my $type = shift;

  if (!defined $type) { $type = '.html'; }

  $url = &Scoop::URLWithoutAnchor ($url);

  # new-style: just use numeric filenames. saves space.
  my $assigned = $self->{page_to_tmpfile}->{$url};
  if (!defined $assigned) {
    my $val = $self->{page_to_tmpfile_acc}++;
    my $name = $self->{outtmp}.$CF::slash.$val.$type;
    $self->add_page_tmpfile ($name, $url);
  }
  $self->{page_to_tmpfile}->{$url};
}

sub href_to_multipage_href {
  my $self = shift;
  my $url = shift;
  my $type = shift;

  my $anchor = &Scoop::URLAnchor ($url);

  if ($CF::debug > 1) {
    my ($class, $file, $line) = caller();
    $self->dbg2 ("may assign page file: $url (at $class:$line)");
  }

  my $fname = $self->href_to_multipage_filename ($url, $type);
  $fname =~ /[\\\/\:]([-_\.A-Za-z0-9]+)$/;
  $1.$anchor;
}

sub href_to_singlepage_href {
  my $self = shift;
  my $url = shift;

  my $anchor = &Scoop::URLAnchor ($url);
  $url = &Scoop::URLWithoutAnchor ($url);
  if ($anchor ne '') {
    $anchor =~ s/\#/__HASH__/g;
    $anchor =~ s/ /_20/g; $anchor =~ s/[^-_A-Za-z0-9]/_/g;
  }

  # new-style: just use numeric anchors. saves space.
  my $assigned = $self->{anchors_assigned}->{$url};
  if (defined $assigned) {
    return $assigned.$anchor;
  } else {
    my $ank = $self->{anchors_assigned_acc}++;
    $self->{anchors_assigned}->{$url} = $ank;
    if ($CF::debug > 1) {
      my ($class, $file, $line) = caller();
      $self->dbg2 ("assigned anchor: $url = #$ank (at $class:$line)");
    }
    return $ank.$anchor;
  }
}

sub got_redirected {
  my ($self, $oldurl, $newurl) = @_;

  $self->{redirections}{$oldurl} = $newurl;
  $self->{redirections_backwards}{$newurl} .= " ".$oldurl;

  if ($CF::fileperpage) {
    my $fname = $self->href_to_multipage_filename($oldurl);
    $self->add_page_tmpfile ($fname, $newurl);
  }
}

sub set_url_title {
  my ($self, $url, $title) = @_;
  if (defined $title && $title =~ /\S/) {
    $title =~ s/\s+/ /g;
    $title =~ s/"/'/g;
    $title =~ s/^ //;
    $title =~ s/ $//;
    $self->{url_title}{$url} = $title;
    push (@{$self->{bookmark_urls}}, $url);
    &Scoop::dbg ("bookmark title for $url: \"$title\".");
  }
}

sub grep_unseen_urls_impl {
  my ($self, $absurl, @links) = @_;
  local ($_);
  my @deduped = ();

  foreach $_ (@links) {
    if (defined $absurl) { $_ = &Scoop::AbsoluteURL ($absurl, $_); }
    next if (defined $self->{already_tagged_for_downloading}{$_});
    $self->{already_tagged_for_downloading}{$_} = 1;
    push (@deduped, $_);
  }

  @deduped;
}

sub grep_unseen_urls {
  my ($self, @links) = @_;
  $self->grep_unseen_urls_impl (undef, @links);
}

sub absolutify_and_grep_unseen_urls {
  my ($self, $url, @links) = @_;
  $self->grep_unseen_urls_impl ($url, @links);
}

sub sitewarn {
  my $self = shift;
  Scoop::sitewarn_file_line ($self->{scf}->{site_defined_at}, @_);
}

sub writing_doc {
  my ($self) = @_;
  ($self->{output_style} == $Scoop::OUT_DOC);
}

sub writing_html {
  my ($self) = @_;
  ($self->{output_style} == $Scoop::OUT_HTML);
}

sub writing_text {
  my ($self) = @_;
  ($self->{output_style} == $Scoop::OUT_TEXT);
}

sub dbg {
  my $self = shift;
  Scoop::dbg(@_);
}

sub dbg2 {
  my $self = shift;
  if ($CF::debug > 1) { Scoop::dbg(@_); }
}

1;


